//*****************************************************************************
//*
//*
//*		FtpDemo.cpp
//*
//*
//*****************************************************************************
#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>	    
#include	"FtpServer.h"
#include	"System/System.h"


//*****************************************************************************
//*
//*		MyFileCheckerCallback
//*
//*****************************************************************************
//	Is called for every file access and directory enumeration
//	hConn		: handle of the connection.
//	pPath		: pointer to the filepath  (i.e. "C:\Temp" )
//	pName		: pointer to the filename  (i.e. "Test.txt")
//	pParam		: is the user parameter of the callback
//	iShare		: is the number of the used share
//	Returns the access right for this file
//					0 = all allowed
//					1 = read only access
//					2 = no allowed but file will be listed
//					3 = the user can't see this file
int _cdecl MyFileCheckerCallback(FtpConnection hConn,const char *pPath,const char *pName,void *pParam,int iShare)
{
const char *pUser;
int			iLen;


	pUser = (const char*)pParam;							// in the parameter we have stored the user name
	iLen  = strlen(pName);

	if(iLen>4 && !strcmp(pName+iLen-4,".obj"))				// hide files which ends with ".obj"
		{
		return 3;
		}

	if(iShare==0 && !strcmp(pName,"Debug"))					// no acces to the debug directory
		{
		return 2;
		}


return 0;
}

//*****************************************************************************
//*
//*		MyLoginCallback
//*
//*****************************************************************************
//	Is called if a user was logged on.
//	hConn		: handle of the connection
//	pUser		: is the user name
//	pPassword	: is the password
//	Returns an error code:
//					0 = wrong username or password
//					1 = we need a password for the user
//					2 = login ok
int _cdecl MyLoginCallback(FtpConnection hConn,const char *pUser,const char *pPassword)
{


	if(!strcmp(pUser,"anonymous"))							// anonymous login with sConfig.cFilePath
		{
		return 2;
		}

	if(*pPassword==0)										// did we have a password
		{
		return 1;
		}

	if(!strcmp(pUser,"smith") && !strcmp(pPassword,"smith_pw"))
		{
		FtpInsertTreeEntry(hConn,"Data"    ,"C:\\Users\\Smith",0,MyFileCheckerCallback,"smith");
		FtpInsertTreeEntry(hConn,"ReadOnly","C:\\Users\\Smith",1,MyFileCheckerCallback,"smith");
		return 2;
		}

	if(!strcmp(pUser,"adams") && !strcmp(pPassword,"adams_pw"))
		{
		FtpInsertTreeEntry(hConn,"Data"    ,"C:\\Users\\Adams",0,MyFileCheckerCallback,"adams");
		FtpInsertTreeEntry(hConn,"ReadOnly","C:\\Users\\Adams",1,MyFileCheckerCallback,"adams");
		return 2;
		}



return 0;
}


//*****************************************************************************
//*
//*		MyConnectCallback
//*
//*****************************************************************************
//	Is called if a new connection to the ftp server was opened.
//	hConn		: handle of the connection.
//	pIpAddr		: ip address of the connected client.
//	usPort		: tcp port of the connected client.
//	Returns 1 if the connection was accepted or 0 if not.
int _cdecl MyConnectCallback(FtpConnection hConn,const char *pIpAddr,unsigned short usPort)
{


	if(!strncmp(pIpAddr,"192.168.17.",11))					// don't allow from "192.168.17.*"
		{
		printf("not accept connection from %s:%u\n",pIpAddr,usPort);
		return 0;
		}


	printf("accept connection from %s:%u\n",pIpAddr,usPort);


return 1;
}

//*****************************************************************************
//*
//*		MyConnectCallback
//*
//*****************************************************************************
//	Is called if a new connection to the ftp server was opened.
//	hConn		: handle of the connection.
//	pIpAddr		: ip address of the connected client.
//	usPort		: tcp port of the connected client.
void _cdecl MyDisconnectCallback(FtpConnection hConn,const char *pIpAddr,unsigned short usPort)
{

	printf("close connection from %s:%u\n",pIpAddr,usPort);


}

//*****************************************************************************
//*
//*		main
//*
//*****************************************************************************
int	main(int iArgc,char **pArgv)
{
const char		   *pIpAddr="0.0.0.0";
const char		   *pMyPath="C:\\Temp";
const char		   *pParam;
FtpConfigData		sConfig;
int					iPort=0;
int					iKey;
int					iRet;
int					iNum;



    for(iNum=1;iNum<iArgc;iNum++)               // scan command line
        {
        pParam = pArgv[iNum];

        if(!strcmp(pParam,"-help") || !strcmp(pParam,"?"))
            {
            printf("\n\n");
            printf("\tFtpDemo   [-i ipaddress] [-p path] [-t tcpport]\n\n");
            printf("\n\n");
            return -2;
            }

        if(!strcmp(pParam,"-i") && iNum+1<iArgc)
            {
            iNum++;
            pIpAddr = pArgv[iNum];
            continue;
            }

        if(!strcmp(pParam,"-p") && iNum+1<iArgc)
            {
            iNum++;
            pMyPath = pArgv[iNum];
            continue;
            }

        if(!strcmp(pParam,"-t") && iNum+1<iArgc)
            {
            iNum++;
            iPort = atoi(pArgv[iNum]);
            continue;
            }
        }


//*****************************************************************************



	memset(&sConfig,0,sizeof(sConfig));

	strcpy(sConfig.cIpAddress,pIpAddr);
	strcpy(sConfig.cFilePath ,pMyPath);
	sConfig.iMaxConnections	= 4;
    sConfig.usTcpPort		= (unsigned short)iPort;
	sConfig.uTimeout		= 1000*60*2;
    sConfig.pCbLogin        = MyLoginCallback;
    sConfig.pCbConnect      = MyConnectCallback;
    sConfig.pCbDisconnect   = MyDisconnectCallback;
	sConfig.pCbChecker		= MyFileCheckerCallback;
	sConfig.pCbCheckerParam	= "anonymous";

	iRet = FtpDaemon(&sConfig);

	if(iRet!=FTP_ERR_NONE)
		{
		switch(iRet)
			{
		case FTP_ERR_SOCKET:		printf("socket error\n");					break;
		case FTP_ERR_ISRUNNING:   	printf("ftp server is allready running\n");	break;
		case FTP_ERR_ISNOTRUNNING:	printf("ftp server is not running\n");		break;
		case FTP_ERR_WRONGIP:     	printf("the ip address is wrong\n");		break;
		case FTP_ERR_INIT:		 	printf("initialization error\n");			break;
		case FTP_ERR_WRONGPATH:		printf("wrong path name\n");				break;
		default:					printf("unknown erro %i\n",iRet);			break;
			}

		return -1;
		}

    printf("\n\nshutdown server with  ESC+ENTER \n\n\n");

	for(;;)
		{
           iKey =getchar();
        if(iKey==27)break;
		if(iKey==10)break;
		}


    printf("\nshutdown server\n");

	FtpStopDaemon();

    printf("\n...exit...\n\n");


return 0;
}

