//*****************************************************************************
//*
//*
//*     FtpInternal.h
//*
//*
//*****************************************************************************
#ifndef     __FTPINTERNAL_H__
#define     __FTPINTERNAL_H__

#include    "System/SocketUtility.h"
#include    "System/System.h"
#include    "FtpServer.h"

#ifndef     FTP_MEMORY_MAX
#define     FTP_MEMORY_MAX  0x10000                         // Max size of FTP heap
#endif
#define     FTP_STACK_SIZE  0x6000                          // Stacksize for the connection thread
#define     FTP_WDOG_TIME   30000                           // Watchdog time out for threads
#define     FTP_VERSION     "AzFtp 1.12"                    // Version string
#define     FTP_PASSIVE_MIN 17318                           // start port for passive connections
#define     FTP_PASSIVE_MAX 19957                           // end port for passive connections
#define     FTP_RECEIVE_T0  10000                           // timeout for data receiving
#define     FTP_LISTEN_T0   10000                           // timeout for listen connections
#define     FTP_MAX_LINE    1024                            // max line size
#ifndef     FTP_EN_CHECK_CB
#define     FTP_EN_CHECK_CB 1                               // enables the check callback
#endif
#ifndef     FTP_MAX_PATH
#define     FTP_MAX_PATH    127                             // size of a share path
#endif
#ifndef     FTP_MAX_NAME
#define     FTP_MAX_NAME    32                              // size of a share name
#endif
#ifndef     FTP_PORT
#define     FTP_PORT        21                              // tcp port number of the server
#endif
#ifndef     FTP_DATA_PORT
#define     FTP_DATA_PORT   20                              // tcp port number of the data connection
#endif
#ifndef     FTP_USER_PARAMS
#define     FTP_USER_PARAMS 4                               // max count of user params
#endif
#ifndef     FTP_SHARES
#define     FTP_SHARES      8                               // max count of shares
#endif

#ifdef      linux
#ifndef     stricmp
#define     stricmp         strcasecmp
#endif
#ifndef     strnicmp
#define     strnicmp        strncasecmp
#endif
#endif


typedef struct
    {
    int                 iNameLen;                           // is the length of the name
    int                 iPathLen;                           // is the length of the path
    char                cName[FTP_MAX_NAME];                // is the name of the share ("" means root entry)
    char                cPath[FTP_MAX_PATH];                // is the path of the share
    char                cReadOnly;                          // read only flag
    FtpFileCheckerCB    pCbChecker;                         // callback for file checking
    void               *pCbParam;                           // parameter for file checking
    }FtpShare;

typedef struct
    {
    TcpHandle           hServer;                            // socket handle for server
    SysSemaphore        hSemaphore;                         // semapore for the server
    unsigned short      usTcpPort;                          // is the tcp port for the server
    unsigned short      usPassivePort;                      // next passive port
    int                 iMaxConnections;                    // maximum count of client connections
    int volatile        iCurConnections;                    // current count of client connections
    int volatile        iShutdown;                          // shutdown the server
    int volatile        iRunning;                           // is the server running
    int                 iPriority;                          // taks priority 0-256
    int                 iReadOnly;                          // is the path read only
    int                 iNoAnonym;                          // don't allow an anonymous login
    unsigned            uTimeout;                           // receive timeout in ms
    char                cFilePath[256];                     // default file path
    char                cIpAddress[64];                     // ip addres off the http-server
    FtpLoginCB          pCbLogin;                           // callback for a client login
    FtpConnectCB        pCbConnect;                         // callback for client connection
    FtpDisconnectCB     pCbDisconnect;                      // callback for client disconnection
    FtpFileCheckerCB    pCbChecker;                         // callback for file checking
    void               *pCbCheckerParam;                    // parameter for file checking
    }FtpServerData;

typedef struct
    {
    FtpServerData      *pServer;                            // pointer to the server data
    TcpHandle           hConnection;                        // socket handle for input
    TcpHandle           hDataConn;                          // handle to data connection
    TcpHandle           hListenConn;                        // handle to passive listen connection
    SysWdog             hWatchdog;                          // wachdog handel fr connection thread
    void               *pUserParams[FTP_USER_PARAMS];       // user parameter
    char                cMovePath[256];                     // file path for renaming
    char                cConnIp   [64];                     // ip addres of the client
    char                cDataIp   [64];                     // ip adress of the data port
    char                cUsername [32];                     // is the current user
    char                cPassword [32];                     // is the current user
    unsigned short      usDataPort;                         // number of the data port
    unsigned short      usConnPort;                         // tcp port of the client
    unsigned short      usPassivePort;                      // tcp port of the active connection
    char                cLineBuffer[FTP_MAX_LINE+4];        // input buffer
    unsigned            uBufferSize;                        // count of the received chars
    unsigned            uBufferPos;                         // reading position in the buffer
    FtpShare            aShares[FTP_SHARES];                // array
    int                 iShareCount;                        // count of used shares
    int                 iShareRoot;                         // is the root share (-1=none)
    int                 iShareSel;                          // is the selected share (-1=none)
    char                cSharePwd[256];                     // working directory depent to share
    int                 iSharePwdLen;                       // length of working directory
    int                 iPathRoot;                          // is the last created path the root path
    int                 iPathLen;                           // length of the last created path
    char                cDataMode;                          // 0=none 1=passive 2=active
    char                cLoginDone;                         // is a user logged in
    char                cAscii;                             // is the type on ACSCII mode
    }FtpData;

//*****************************************************************************
//*
//*     MallocFunctions
//*
//*****************************************************************************

extern        void *hFtpHeap;
static inline void *FtpAlloc(unsigned uSize){return SysHeapAlloc(hFtpHeap,uSize);}
static inline void  FtpFree (void    *pPtr ){       SysHeapFree (hFtpHeap,pPtr );}


#endif
