//*****************************************************************************
//*
//*
//*     FtpServer.cpp
//*
//*
//*****************************************************************************
#include    <stdio.h>
#include    <stdlib.h>
#include    <string.h>
#include    "FtpServer.h"
#include    "FtpInternal.h"
#include    "System/System.h"
#include    "System/SocketUtility.h"

#ifndef		LOG_RECV_ON
#define     LOG_RECV_ON     0
#endif
#ifndef		LOG_SEND_ON
#define     LOG_SEND_ON     0
#endif
#ifndef		LOG_LIST_ON
#define     LOG_LIST_ON     0
#endif
#ifndef		LOG_CONN_ON
#define     LOG_CONN_ON     0
#endif

#if         LOG_RECV_ON
#define     LOG_RECV(p,l)   printf("RECV:  %.*s\n",l,p);
#else
#define     LOG_RECV(p,l)
#endif

#if         LOG_SEND_ON
#define     LOG_SEND(p,l)   printf("SEND:  %.*s\n",(l<2 || (p)[(l)-2]!='\r' || (p)[(l)-1]!='\n')? l:l-2,p);
#else
#define     LOG_SEND(p,l)
#endif

#if         LOG_LIST_ON
#define     LOG_LIST(p,l)   printf("LIST:  %.*s\n",(l<2 || (p)[(l)-2]!='\r' || (p)[(l)-1]!='\n')? l:l-2,p);
#else
#define     LOG_LIST(p,l)
#endif

#if         LOG_CONN_ON
#define     LOG_CONN(i,p,c) {char cAddr[64];    IpToString(&i,cAddr); printf("CONN:  %s from:%s:%u\n",c,cAddr,p);}
#else
#define     LOG_CONN(i,p,c)
#endif

#if			FTP_EN_CHECK_CB
#define 	FTP_CHECKER(a)	a
#else
#define 	FTP_CHECKER(a)	0
#endif


        void           *hFtpHeap        = 0;                // heap for the http server
static  int volatile    iServerRunning  = 0;                // is the daemon started
static  FtpServerData   sFtpServer;                         // instance of the http server
static  const char     *pMonthNames[16]=
                            {
                            "Jan","Jan","Feb","Mar",
                            "Apr","May","Jun","Jul",
                            "Aug","Seb","Oct","Nov",
                            "Dec","Dec","Dec","Dec"
                            };



//*****************************************************************************
//*
//*     AllocFtpData
//*
//*****************************************************************************
//  Allocates the data pointer for a new client connection
//  Returns the data pointer or zero if an error occurs.
static FtpData *AllocFtpData()
{
FtpData *pData;


        pData = (FtpData*)SysHeapAlloc(hFtpHeap,sizeof(FtpData));
    if(!pData)return 0;


    memset(pData,0,sizeof(FtpData));


return pData;
}

//*****************************************************************************
//*
//*     FreeFtpData
//*
//*****************************************************************************
//  Deallocates a data pointer from a new client connection
static void FreeFtpData(FtpData *pData)
{

    if(!pData)return;

    SysHeapFree(hFtpHeap,pData);


}

//*****************************************************************************
//*
//*     FtpCloseListenPort
//*
//*****************************************************************************
//  Closes a open listen port for the pssive connection
//  pData       : pointer to the server data
static void FtpCloseListenPort(FtpData *pData)
{

    if(pData->hListenConn)
        {
        TcpClose(pData->hListenConn);
        pData->hListenConn=0;
        }

}

//*****************************************************************************
//*
//*     FtpOpenListenPort
//*
//*****************************************************************************
//  Opens a listen port for the pssive connection
//  pData       : pointer to the server data
//  Returns 1 if succeful or 0 if an error occurs
static int FtpOpenListenPort(FtpData *pData)
{
IpAddr      iAddr;
int         iTurn;


    SysThreadCheck();


    if(pData->hListenConn)
        {
        return 1;
        }

    if(!IpAddress(&iAddr,pData->pServer->cIpAddress))
        {
        return 0;
        }

    for(iTurn=0;iTurn<16;iTurn++)                           // open data listen port
        {
        SysThreadCheck();

        SysSemaphoreLock(sFtpServer.hSemaphore);

        if(pData->pServer->usPassivePort<FTP_PASSIVE_MIN ||
           pData->pServer->usPassivePort>FTP_PASSIVE_MAX)
            {
            pData->pServer->usPassivePort=FTP_PASSIVE_MIN;
            }


        SysThreadCheck();
        SysWatchdogTrigger(pData->hWatchdog);
        SysWatchdogChangeTime(pData->hWatchdog,150000,0);

        pData->hListenConn = TcpOpen(&iAddr,pData->pServer->usPassivePort,TCP_SERVER,0);

        SysWatchdogChangeTime(pData->hWatchdog,FTP_WDOG_TIME,0);
        SysThreadCheck();

        pData->usPassivePort = pData->pServer->usPassivePort;
        pData->pServer->usPassivePort++;

        SysSemaphoreUnlock(sFtpServer.hSemaphore);

        if(pData->hListenConn)break;
        if(iTurn%8==0)SysSleep(100);
        }


    SysThreadCheck();


    if(!pData->hListenConn)
        {
        return 0;
        }


return 1;
}

//*****************************************************************************
//*
//*     FtpOpenDataPort
//*
//*****************************************************************************
//  Opens a tcp connection for the data transver
//  pData       : pointer to the server data
//  Returns a TcpHandle ot 0 if an error occurs
static TcpHandle FtpOpenDataPort(FtpData *pData)
{
TcpHandle       hTcp;
TcpHandle       hTemp;
IpAddr          iAddr;
unsigned        uTurn;
unsigned short  usPort;



    if(pData->cDataMode==2)                                 // active connection
        {
        if(!IpAddress(&iAddr,pData->cDataIp,0))return 0;

        SysThreadCheck();
        SysWatchdogTrigger(pData->hWatchdog);
        SysWatchdogChangeTime(pData->hWatchdog,150000,0);

        hTcp = TcpOpen(&iAddr,pData->usDataPort,0,TCP_REUSEADDR|TCP_USEPORT(FTP_DATA_PORT));

        LOG_CONN(iAddr,pData->usDataPort,"open active connection");

        SysWatchdogChangeTime(pData->hWatchdog,FTP_WDOG_TIME,0);
        SysWatchdogTrigger(pData->hWatchdog);
        SysThreadCheck();

        return hTcp;
        }


    if(pData->cDataMode==1)                                 // passive connection
        {
        SysThreadCheck();
        SysWatchdogTrigger(pData->hWatchdog);


            hTcp = TcpAcceptWait(pData->hListenConn,&iAddr,&usPort,FTP_LISTEN_T0);
        if(!hTcp)
            {
            for(uTurn=0;uTurn<2;uTurn++)
                {
                SysThreadCheck();
                SysWatchdogTrigger(pData->hWatchdog);

                    hTemp = TcpAcceptWait(pData->hListenConn,&iAddr,&usPort,FTP_LISTEN_T0);
                if(!hTemp)break;

                hTcp = hTemp;
                }
            }

        if(!hTcp)                                           // close listen socket at an error
            {
            TcpClose(pData->hListenConn);
            pData->hListenConn = 0;
            }
        else{
            LOG_CONN(iAddr,usPort,"accept passive connection");
            }

        SysWatchdogTrigger(pData->hWatchdog);
        SysThreadCheck();



        return hTcp;
        }


return 0;
}

//*****************************************************************************
//*
//*     CheckPath
//*
//*****************************************************************************
//  Checks all entries of a file path with the file checker callback
//  pData       : pointer to the server data
//  pPath       : pointer to the file path
//  iLen        : length of the file path
//  iPos        : position of first name for checking
//  iShare      : used share number
//  Returns 0 if the path is accessable or 1 if not
static int CheckPath(FtpData *pData,char *pPath,int iLen,int iPos,int iShare)
{
int     iOff;
int     iOld;
int     iNum;
int     iRet;


    for(iOff=iPos;;iOff++)                                  // check all directories in the path
        {
        if(pPath[iOff]!='/' && pPath[iOff]!='\\' && pPath[iOff]!=0)
            {
            continue;
            }

        if(iPos>0){iOld=pPath[iPos-1];pPath[iPos-1]=0;}

        iNum        = pPath[iOff];
        pPath[iOff] = 0;
        iRet        = pData->aShares[iShare].pCbChecker(pData,pPath,pPath+iPos,pData->aShares[iShare].pCbParam,iShare);

        if(iRet>=2)                                         // no acces to this directory
            {
            return 1;
            }

        iPos        = iOff+1;
        pPath[iOff] = iNum;

        if(iPos>0)pPath[iPos-1] = iOld;
        if(iOff>=iLen)break;
        }


return 0;
}

//*****************************************************************************
//*
//*     CreatePathName
//*
//*****************************************************************************
//  Creates a full Pathname
//  pData       : pointer to the server data
//  pName       : input path
//  pPath       : pointer to the buffer for the new path
//  iSize       : is the size of the buffer
//  iMode       : is the access mode (0=read 1=write 2=no check)
//  Return the used share or -1 in an error occurs.
static int CreatePathName(FtpData *pData,const char *pName,char *pPath,int iSize,int iMode=0)
{
int     iShare;
int     iNull;
int     iRoot;
int     iPos;
int     iLen;
int     iMin;
int     iAdd;
int     iNum;
int     iRet;


    pPath[0]        = 0;
    pData->iPathLen = 0;
    pData->iPathRoot= 0;


    if(pName[0]=='/' || pName[0]=='\\')                     // is it an absolute path
        {
        pName +=  1;
        iShare = -1;
        iAdd   =  1;
        }
    else{
        iShare =  pData->iShareSel;
        iAdd   =  0;
        }


    if(iShare<0)
        {
        for(iLen=0;pName[iLen];iLen++)                      // length of share name
            {
            if(pName[iLen]=='/' )break;
            if(pName[iLen]=='\\')break;
            }

        iNull = -1;

        for(iShare=pData->iShareCount-1;iShare>=0;iShare--) // find used share
            {
            if(pData->aShares[iShare].iNameLen==0)iNull=iShare;
            if(pData->aShares[iShare].iNameLen!=iLen)continue;
            if(memcmp(pData->aShares[iShare].cName,pName,iLen))continue;

            pName += iLen;

            if(pName[0]=='/' || pName[0]=='\\')
                {
                pName++;
                }

            break;
            }


        if(iShare<0)                                        // no share found
            {
            if(iNull<0)
                {
                if(!pName[0] || (pName[0]=='.' && pName[1]==0) || (pName[0]=='.' && (pName[1]=='/' || pName[1]=='\\') && pName[2]==0))
                    {
                    pData->iPathRoot=1;
                    }

                return -1;
                }

            iShare = iNull;
            }
        }


    iSize -= 1;
    iLen   = pData->aShares[iShare].iPathLen;


    if(iLen>=iSize)
        {
        return -1;
        }

    memcpy(pPath,pData->aShares[iShare].cPath,iLen);        // copy base path
    iMin = iLen-1;

    if(!iAdd && pData->iSharePwdLen)
        {
        if(iLen+pData->iSharePwdLen+1>=iSize)return -1;
        memcpy(pPath+iLen,pData->cSharePwd,pData->iSharePwdLen);
        iLen += pData->iSharePwdLen;
        pPath[iLen] = '/';
        iLen++;
        iRoot = 0;
        }
    else{
        iRoot = (pData->aShares[iShare].iNameLen)? 0:1;
        }


    for(;iLen<=iSize;iLen++,pName++)                        // decode path
        {
        if(iRoot && iLen>iMin+1)                            // jump from root to other share
        if(pName[0]==0 || pName[0]=='/' || pName[0]=='\\')
            {
            iNum = iLen-iMin-1;

            for(iPos=pData->iShareCount-1;iPos>=0;iPos--)
                {
                if(pData->aShares[iPos].iNameLen!=iNum)continue;
                if(memcmp(pData->aShares[iPos].cName,pPath+iMin+1,iNum))continue;

                if(iLen>=iSize)return -1;

                iLen   = pData->aShares[iPos].iPathLen-1;
                memcpy(pPath,pData->aShares[iPos].cPath,iLen);
                iMin   = iLen;
                iShare = iPos;
                break;
                }

            iRoot = 0;
            }

           pPath[iLen] = *pName;
        if(pPath[iLen]==0)                                  // end of input name
            {
            if(iShare<0 || iShare==pData->iShareRoot)
                {
                if(iLen<=iMin+1)
                    {
                    pData->iPathRoot = 1;
                    }

                if(iShare<0)
                    {
                    pPath[0] = 0;
                    iLen     = 0;
                    }
                }

            pData->iPathLen = iLen;

                                                            // exists a file checker callback
            if(FTP_CHECKER(pData->aShares[iShare].pCbChecker) && iMode!=2 && iLen>0)
                {
                for(iLen--;iLen>0;iLen--)
                    {
                    if(pPath[iLen]=='/' )break;
                    if(pPath[iLen]=='\\')break;
                    }

                iNull       = pPath[iLen];
                pPath[iLen] = 0;
                iRet        = pData->aShares[iShare].pCbChecker(pData,pPath,pPath+iLen+1,pData->aShares[iShare].pCbParam,iShare);
                pPath[iLen] = iNull;

                if(iRet>=3 || (iRet>=1 && iMode))           // no acces to this file
                    {
                    return -1;
                    }
                }


            return iShare;
            }

        if(pPath[iLen]=='\\')                               // change \ to /
            {
            pPath[iLen]='/';
            continue;
            }

        if(pPath[iLen]!='.')continue;                       // "." current directory

        if(pName[1]=='/' || pName[1]=='\\' || pName[1]==0)
            {
            for(iLen--;iLen>=0;iLen--)
                {
                if(pPath[iLen]=='/')break;
                }

            if(iLen<iMin || (iLen==iMin && !*pName))
                {
                if(iShare<0 || iShare==pData->iShareRoot)
                    {
                    break;
                    }
                }

            iLen--;
            continue;
            }

        if(pName[1]!='.')continue;                          // upper directory

        if(pName[2]=='/' || pName[2]=='\\' || pName[2]==0)
            {
            pName+=2;

            for(iLen--;iLen>=0;iLen--)
                {
                if(pPath[iLen]=='/')break;
                }

            for(iLen--;iLen>=0;iLen--)
                {
                if(pPath[iLen]=='/')break;
                }


            if(iLen<iMin)                                   // jump to root share
                {
                if(iShare<0 || iShare==pData->iShareRoot)
                    {
                    break;
                    }

                   iShare = pData->iShareRoot;
                if(iShare>=0)                               // copy base path
                    {
                    iLen = pData->aShares[iShare].iPathLen;
                    if(iLen>=iSize)return -1;
                    memcpy(pPath,pData->aShares[iShare].cPath,iLen);
                    iLen--;
                    iMin = iLen;
                    }
                else{                                       // create dummy path
                    pPath[0] = '/';
                    pPath[1] = 0;
                    iMin     = 0;
                    iLen     = 0;
                    }

                iRoot = 1;
                }

            if(!pName[0])
                {
                pName--;
                iLen--;
                }
            }
        }


    pPath[0] = 0;


return -1;
}


//*****************************************************************************
//*
//*     FtpGetLine
//*
//*****************************************************************************
//  Receives the next line terminated with "\r\n"
//  pData       : pointer to the server data
//  pBuffer     : here the buffer of the line will be stored
//  uTimeout    : is the receive timeout in ms
//  Returns the size of the line or -1 if an error occurs.
static int FtpGetLine(FtpData *pData,char **pBuffer,unsigned uTimeout)
{
unsigned    uEnd;
int         iNum;
int         iRet;
int         iLen;
int         iPos;
int         iMax;


    iLen = pData->uBufferSize-pData->uBufferPos;

    if(iLen>0)
        {
        memmove(pData->cLineBuffer,pData->cLineBuffer+pData->uBufferPos,iLen);
        pData->uBufferSize = iLen;
        pData->uBufferPos  = 0;
        }
    else{
        pData->uBufferSize = 0;
        pData->uBufferPos  = 0;
        iLen = 0;
        }


    uEnd = SysTickCount()+uTimeout;


    for(iNum=0;;iNum++)
        {
           iRet = TcpGetNext(pData->hConnection,pData->cLineBuffer+iLen,FTP_MAX_LINE-iLen,uTimeout);
        if(iRet<0)
            {
            *pBuffer="";
            return 0;
            }


        *pBuffer = pData->cLineBuffer;
        pData->uBufferSize += iRet;
        iMax = pData->uBufferSize-1;

        for(iPos=0;iPos<iMax;iPos++)
            {
            if(pData->cLineBuffer[iPos  ]!='\r')continue;
            if(pData->cLineBuffer[iPos+1]=='\n')
                {
                pData->cLineBuffer[iPos] = 0;
                pData->uBufferPos = iPos+2;
                return iPos;
                }
            }


        if(iMax>=FTP_MAX_LINE || !uTimeout)
            {
            break;
            }


        iLen     = pData->uBufferSize;
        uTimeout = uEnd-SysTickCount();

        if(uTimeout&0x80000000)
            {
            uTimeout = 0;
            }
        }


    iMax++;
    pData->cLineBuffer[iMax] = 0;
    pData->uBufferPos = iMax;


return iMax;
}

//*****************************************************************************
//*
//*     CmdUser
//*
//*****************************************************************************
//  Handle a USER command.    USER <username>
//  pData       : Ponter to the connection data
//  pLine       : Ponter to the command line
static void CmdUser(FtpData *pData,char *pLine)
{
TcpHandle   hClient;
int         iLen;
int         iRet;



    hClient = pData->hConnection;


    for(iLen=0;iLen<sizeof(pData->cUsername)-1;iLen++)      // copy username
        {
        if(pLine[iLen]>=0 && pLine[iLen]<=' ')break;
        pData->cUsername[iLen] = pLine[iLen];
        }

    pData->cLoginDone      =  0;
    pData->cUsername[iLen] =  0;
    pData->cPassword[0]    =  0;
    pData->cSharePwd[0]    =  0;
    pData->iSharePwdLen    =  0;
    pData->iShareCount     =  0;
    pData->iShareSel       = -1;


    if(pData->pServer->pCbLogin)                            // use callback for login
        {
        iRet = pData->pServer->pCbLogin(pData,pData->cUsername,pData->cPassword);

        if(iRet==0)
            {
            TcpPut(hClient,"530 Not logged in.\r\n",20);
            LOG_SEND(      "530 Not logged in.\r\n",20);
            pData->cUsername[0] = 0;
            return;
            }

        if(iRet==1)
            {
            TcpPut(hClient,"331 User name okay, need password.\r\n",36);
            LOG_SEND(      "331 User name okay, need password.\r\n",36);
            return;
            }

        }
    else{                                                   // only anonymous login
        if(pData->pServer->iNoAnonym || strcmp(pData->cUsername,"anonymous"))
            {
            TcpPut(hClient,"530 Not logged in.\r\n",20);
            LOG_SEND(      "530 Not logged in.\r\n",20);
            pData->cUsername[0] = 0;
            return;
            }
        }


    TcpPut(hClient,"230 User logged in, proceed.\r\n",30);
    LOG_SEND(      "230 User logged in, proceed.\r\n",30);

    pData->cLoginDone = 1;

    if(pData->iShareCount==0)                               // insert default share
        {
        FtpInsertTreeEntry((FtpConnection)pData,"",pData->pServer->cFilePath,pData->pServer->iReadOnly,FTP_CHECKER(pData->pServer->pCbChecker),pData->pServer->pCbCheckerParam);
        pData->iShareSel = 0;
        }

}

//*****************************************************************************
//*
//*     CmdPass
//*
//*****************************************************************************
//  Handle a PASS command.    PASS <password>
//  pData       : Ponter to the connection data
//  pLine       : Ponter to the command line
static void CmdPass(FtpData *pData,char *pLine)
{
TcpHandle   hClient;
int         iPos;
int         iRet;


    hClient = pData->hConnection;


    if(!pData->cUsername[0])                                // username allready set
        {
        TcpPut(hClient,"332 Need account for login.\r\n",29);
        LOG_SEND(      "332 Need account for login.\r\n",29);
        return;
        }


    for(iPos=0;;iPos++)                                     // copy password
        {
        if(pLine[iPos]=='\r' || pLine[iPos]=='\n' || pLine[iPos]==0)
            {
            pData->cPassword[iPos] = 0;
            break;
            }

        if(iPos>=sizeof(pData->cPassword)-1)
            {
            pData->cPassword[0] = 0;
            TcpPut(hClient,"530 Not logged in.\r\n",20);
            LOG_SEND(      "530 Not logged in.\r\n",20);
            return;
            }

        pData->cPassword[iPos] = pLine[iPos];
        }


    if(pData->pServer->pCbLogin)                            // use callback for login
        {
        iRet = pData->pServer->pCbLogin(pData,pData->cUsername,pData->cPassword);

        if(iRet!=2)
            {
            pData->iShareCount     =  0;
            pData->iShareSel       = -1;

            TcpPut(hClient,"530 Not logged in.\r\n",20);
            LOG_SEND(      "530 Not logged in.\r\n",20);
            return;
            }
        }


    if(pData->iShareCount==0)                               // insert default share
        {
        FtpInsertTreeEntry((FtpConnection)pData,"",pData->pServer->cFilePath,pData->pServer->iReadOnly,FTP_CHECKER(pData->pServer->pCbChecker),pData->pServer->pCbCheckerParam);
        pData->iShareSel = 0;
        }


    TcpPut(hClient,"230 User logged in, proceed.\r\n",30);
    LOG_SEND(      "230 User logged in, proceed.\r\n",30);

    pData->cLoginDone = 1;


}

//*****************************************************************************
//*
//*     CmdPasv
//*
//*****************************************************************************
//  Handle a PASV command.    PASV
//  pData       : Ponter to the connection data
static void CmdPasv(FtpData *pData)
{
char        cTemp[128];
char        cAddr[64];
TcpHandle   hClient;
IpAddr      iAddr;
int         iPos;
int         iLen;



    hClient = pData->hConnection;


    if(!FtpOpenListenPort(pData))                           // exists an listen port
        {
        TcpPut(hClient,"421 Service not available, closing control connection.\r\n",56);
        LOG_SEND(      "421 Service not available, closing control connection.\r\n",56);
        return;
        }


    TcpGetLocal(pData->hConnection,&iAddr);                 // get address from the own side
    IpToString(&iAddr,cAddr);


    if(!strcmp(cAddr,"0.0.0.0") || IsIpV6Ex(&iAddr))
        {
        TcpPut(hClient,"421 Service not available, closing control connection.\r\n",56);
        LOG_SEND(      "421 Service not available, closing control connection.\r\n",56);
        return;
        }

    for(iPos=0;cAddr[iPos];iPos++)                          // covert '.' to ','
        {
        if(cAddr[iPos]=='.')cAddr[iPos]=',';
        }

    iLen = sprintf(cTemp,"227 Entering Passive Mode (%s,%i,%i)\r\n",cAddr,pData->usPassivePort/256,pData->usPassivePort%256);

    TcpPut(hClient,cTemp,iLen);
    LOG_SEND(      cTemp,iLen);

    pData->cDataMode = 1;


}

//*****************************************************************************
//*
//*     CmdFeat
//*
//*****************************************************************************
//  Handle a FEAT command.    FEAT
//  pData       : Ponter to the connection data
static void CmdFeat(FtpData *pData)
{
char        cTemp[128];
TcpHandle   hClient;
int         iLen;


    hClient = pData->hConnection;

    iLen = sprintf(cTemp,"211-Extensions supported\r\n MDTM\r\n EPRT\r\n SIZE\r\n EPSV\r\n211 end\r\n");

    TcpPut(hClient,cTemp,iLen);
    LOG_SEND(      cTemp,iLen);



}

//*****************************************************************************
//*
//*     CmdEpsv
//*
//*****************************************************************************
//  Handle a EPSV command.    EPSV
//  pData       : Ponter to the connection data
//  pLine       : Ponter to the command line
static void CmdEpsv(FtpData *pData)
{
char        cTemp[128];
TcpHandle   hClient;
int         iLen;



    hClient = pData->hConnection;


    if(!FtpOpenListenPort(pData))                           // exists an listen port
        {
        TcpPut(hClient,"421 Service not available, closing control connection.\r\n",56);
        LOG_SEND(      "421 Service not available, closing control connection.\r\n",56);
        return;
        }


    iLen = sprintf(cTemp,"229 Entering Extended Passive Mode (|||%u|).\r\n",pData->usPassivePort);

    TcpPut(hClient,cTemp,iLen);
    LOG_SEND(      cTemp,iLen);

    pData->cDataMode = 1;



}

//*****************************************************************************
//*
//*     CmdPwd
//*
//*****************************************************************************
//  Handle a PWD command.    PWD
//  pData       : Ponter to the connection data
static void CmdPwd(FtpData *pData)
{
char       *pName;
char        cTemp[320];
TcpHandle   hClient;
int         iShare;
int         iLen;


    hClient = pData->hConnection;


    if(!pData->cLoginDone)                                  // username allready set
        {
        TcpPut(hClient,"530 Not logged in.\r\n",20);
        LOG_SEND(      "530 Not logged in.\r\n",20);
        return;
        }

    iShare = pData->iShareSel;

    if(iShare<0)
        {
        iLen = sprintf(cTemp,"257 \"/\" is current directory.\r\n");
        }
    else if(pData->cSharePwd[0]==0)
        {
        iLen = sprintf(cTemp,"257 \"/%s\" is current directory.\r\n",pData->aShares[iShare].cName);
        }
    else{
        pName = pData->aShares[iShare].cName;

        if(*pName)
                iLen  = sprintf(cTemp,"257 \"/%s/%s\" is current directory.\r\n",pName,pData->cSharePwd);
        else    iLen  = sprintf(cTemp,"257 \"/%s\" is current directory.\r\n"   ,pData->cSharePwd);
        }


    TcpPut(hClient,cTemp,iLen);
    LOG_SEND(      cTemp,iLen);

}

//*****************************************************************************
//*
//*     CmdCwd
//*
//*****************************************************************************
//  Handle a CWD command.    CWD <path>
//  pData       : Ponter to the connection data
//  pLine       : Ponter to the command line
static void CmdCwd(FtpData *pData,char *pLine)
{
char        cPath[320];
TcpHandle   hClient;
int         iShare;
int         iAttr;
int         iLen;
int         iPos;




    hClient = pData->hConnection;


    if(!pData->cLoginDone)                              // username allready set
        {
        TcpPut(hClient,"530 Not logged in.\r\n",20);
        LOG_SEND(      "530 Not logged in.\r\n",20);
        return;
        }


       iShare  = CreatePathName(pData,pLine,cPath,sizeof(cPath)-2,2);
    if(iShare<0)
        {
        if(!pData->iPathRoot)
            {
            TcpPut(hClient,"550 Requested action not taken.\r\n",33);
            LOG_SEND(      "550 Requested action not taken.\r\n",33);
            return;
            }

        pData->iShareSel    = -1;
        pData->iSharePwdLen =  0;
        pData->cSharePwd[0] =  0;
        }
    else{
        iPos  = pData->aShares[iShare].iPathLen;
        iLen  = pData->iPathLen-iPos;

        if(pData->iPathLen>0 && cPath[pData->iPathLen-1]=='/')
            {
            cPath[pData->iPathLen-1] = 0;
            iLen--;
            }

        if(iLen>=(int)sizeof(pData->cSharePwd))
            {
            TcpPut(hClient,"550 Requested action not taken.\r\n",33);
            LOG_SEND(      "550 Requested action not taken.\r\n",33);
            return;
            }

        if(iLen>0)
            {
               iAttr = SysGetAttributes(cPath);
            if(iAttr<0 || !(iAttr&SYS_DIR))             // exists the path
                {
                TcpPut(hClient,"550 Requested action not taken.\r\n",33);
                LOG_SEND(      "550 Requested action not taken.\r\n",33);
                return;
                }
            }
        else if(iLen<0)
            {
            iLen = 0;
            }
                                                        // exists a file checker callback
        if(FTP_CHECKER(pData->aShares[iShare].pCbChecker) && iLen>iPos)
            {
            if(CheckPath(pData,cPath,iPos+iLen,iPos,iShare))
                {
                TcpPut(hClient,"550 Requested action not taken.\r\n",33);
                LOG_SEND(      "550 Requested action not taken.\r\n",33);
                return;
                }
            }


        pData->iShareSel        = iShare;               // save new working directory
        pData->iSharePwdLen     = iLen;
        pData->cSharePwd[iLen]  = 0;
        memcpy(pData->cSharePwd,cPath+iPos,iLen);
        }


    TcpPut(hClient,"250 CWD command successful.\r\n",29);
    LOG_SEND(      "250 CWD command successful.\r\n",29);


}

//*****************************************************************************
//*
//*     CmdCdup
//*
//*****************************************************************************
//  Handle a CDUP command.    CDUP
//  pData       : Ponter to the connection data
static void CmdCdup(FtpData *pData)
{

    CmdCwd(pData,"..");

}


//*****************************************************************************
//*
//*     CmdEprt
//*
//*****************************************************************************
//  Handle a EPRT command.    EPRT |<type>|<address>|<port>
//  pData       : Ponter to the connection data
//  pLine       : Ponter to the command line
static void CmdEprt(FtpData *pData,char *pLine)
{
TcpHandle   hClient;
int         iPort;
int         iPos;
int         iVal;



    hClient = pData->hConnection;


    if(pLine[0]!='|' || pLine[2]!='|' || pLine[1]>'9' || pLine[1]<'0')
        {
        TcpPut(hClient,"500 Syntax error, command unrecognized.\r\n",41);
        LOG_SEND(      "500 Syntax error, command unrecognized.\r\n",41);
        return;
        }



    iVal = pLine[1]-'0';                                    // check address type

    #if SU_USE_IPV6
    if(iVal<=0 || iVal>2)
    #else
    if(iVal!=1)
    #endif
        {
        TcpPut(hClient,"522 Extended Port Failure - unknown network protocol.\r\n",55);
        LOG_SEND(      "522 Extended Port Failure - unknown network protocol.\r\n",55);
        return;
        }

    pLine += 3;
    iPort  = 0;

    for(iPos=0;iPos<sizeof(pData->cDataIp)-1;iPos++)        // find end of address
        {
        if(pLine[iPos]== 0 )break;
        if(pLine[iPos]=='|')
            {
            iPort = atoi(pLine+iPos+1);
            break;
            }
        }


    if(iPort<=0 || iPort>=0xFFFF)                           // check port address
        {
        TcpPut(hClient,"500 Syntax error, command unrecognized.\r\n",41);
        LOG_SEND(      "500 Syntax error, command unrecognized.\r\n",41);
        return;
        }

    memcpy(pData->cDataIp,pLine,iPos);
    pData->cDataIp[iPos] = 0;
    pData->usDataPort    = (unsigned short)iPort;
    pData->cDataMode     = 2;


    TcpPut(hClient,"200 command successful.\r\n",25);
    LOG_SEND(      "200 command successful.\r\n",25);


}

//*****************************************************************************
//*
//*     CmdPort
//*
//*****************************************************************************
//  Handle a PORT command.    PORT <ip>,<ip>,<ip>,<ip>,<port>,<port>
//  pData       : Ponter to the connection data
//  pLine       : Ponter to the command line
static void CmdPort(FtpData *pData,char *pLine)
{
char        cTemp[64]="|1|";
TcpHandle   hClient;
int         iPort;
int         iPos;
int         iNum;
int         iEnd;



    hClient = pData->hConnection;



    for(iPos=0,iNum=0;iPos<6*4;iPos++)                      // convert to ip address
        {
        if(pLine[iPos]==0)
            {
            break;
            }

        if(pLine[iPos]==',')
            {
            iNum++;
            if(iNum==4)break;
            cTemp[3+iPos] = '.';
            }
        else{
            cTemp[3+iPos] = pLine[iPos];
            }
        }

    if(pLine[iPos]!=',')
        {
        TcpPut(hClient,"500 Syntax error, command unrecognized.\r\n",41);
        LOG_SEND(      "500 Syntax error, command unrecognized.\r\n",41);
        return;
        }


    iEnd  = iPos+3;
    iPos += 1;
    iPort = atoi(pLine+iPos)*256;                           // high byte of tcp port


    for(;iPos<6*4;iPos++)
        {
        if(pLine[iPos]==0)
            {
            break;
            }

        if(pLine[iPos]==',')
            {
            break;
            }
        }


    if(pLine[iPos]!=',')
        {
        TcpPut(hClient,"500 Syntax error, command unrecognized.\r\n",41);
        LOG_SEND(      "500 Syntax error, command unrecognized.\r\n",41);
        return;
        }


    iPort += atoi(pLine+iPos+1);                            // low byte of tcp port

    sprintf(cTemp+iEnd,"|%i|",iPort);
    CmdEprt(pData,cTemp);

}

//*****************************************************************************
//*
//*     CmdMode
//*
//*****************************************************************************
//  Handle a MODE command.    MODE <mode S,B,C>
//  pData       : Ponter to the connection data
//  pLine       : Ponter to the command line
static void CmdMode(FtpData *pData,char *pLine)
{
TcpHandle   hClient;


    hClient = pData->hConnection;

    if(pLine[0]!='S')
        {
        TcpPut(hClient,"504 Command not implemented for that parameter.\r\n",49);
        LOG_SEND(      "504 Command not implemented for that parameter.\r\n",49);
        return;
        }

    TcpPut(hClient,"200 command successful.\r\n",25);
    LOG_SEND(      "200 command successful.\r\n",25);


}

//*****************************************************************************
//*
//*     CmdType
//*
//*****************************************************************************
//  Handle a TYPE command.    TYPE <type A,E,I,L>
//  pData       : Ponter to the connection data
//  pLine       : Ponter to the command line
static void CmdType(FtpData *pData,char *pLine)
{
TcpHandle   hClient;


    hClient = pData->hConnection;

    if(pLine[0]!='I' && pLine[0]!='A')
        {
        TcpPut(hClient,"504 Command not implemented for that parameter.\r\n",49);
        LOG_SEND(      "504 Command not implemented for that parameter.\r\n",49);
        return;
        }

    pData->cAscii = (pLine[0]=='A')? 1:0;

    TcpPut(hClient,"200 command successful.\r\n",25);
    LOG_SEND(      "200 command successful.\r\n",25);


}

//*****************************************************************************
//*
//*     CmdStru
//*
//*****************************************************************************
//  Handle a STRU command.    STRU <type F,R,P>
//  pData       : Ponter to the connection data
//  pLine       : Ponter to the command line
static void CmdStru(FtpData *pData,char *pLine)
{
TcpHandle   hClient;


    hClient = pData->hConnection;

    if(pLine[0]!='F')
        {
        TcpPut(hClient,"504 Command not implemented for that parameter.\r\n",49);
        LOG_SEND(      "504 Command not implemented for that parameter.\r\n",49);
        return;
        }

    TcpPut(hClient,"200 command successful.\r\n",25);
    LOG_SEND(      "200 command successful.\r\n",25);


}

//*****************************************************************************
//*
//*     CmdRnfr
//*
//*****************************************************************************
//  Handle a RNFR command.    RNFR <path>
//  pData       : Ponter to the connection data
//  pLine       : Ponter to the command line
static void CmdRnfr(FtpData *pData,char *pLine)
{
TcpHandle   hClient;
int         iShare;
int         iAttr;



    hClient = pData->hConnection;

    if(!pData->cLoginDone)                              // username allready set
        {
        TcpPut(hClient,"530 Not logged in.\r\n",20);
        LOG_SEND(      "530 Not logged in.\r\n",20);
        return;
        }

       iShare = CreatePathName(pData,pLine,pData->cMovePath,sizeof(pData->cMovePath));
    if(iShare<0 || (pData->iPathLen>0 && pData->cMovePath[pData->iPathLen-1]=='/') || pData->iPathLen<=pData->aShares[iShare].iPathLen)
        {
        TcpPut(hClient,"553 Requested action not taken.\r\n",33);
        LOG_SEND(      "553 Requested action not taken.\r\n",33);
        pData->cMovePath[0]=0;
        return;
        }

    if(pData->aShares[iShare].cReadOnly)                // is it a read oly path
        {
        TcpPut(hClient,"550 Requested action not taken.\r\n",33);
        LOG_SEND(      "550 Requested action not taken.\r\n",33);
        pData->cMovePath[0]=0;
        return;
        }

       iAttr = SysGetAttributes(pData->cMovePath);      // exist the file
    if(iAttr<0)
        {
        TcpPut(hClient,"450 Requested file action not taken.\r\n",38);
        LOG_SEND(      "450 Requested file action not taken.\r\n",38);
        pData->cMovePath[0]=0;
        return;
        }

    TcpPut(hClient,"350 File or dirctory exist, ready for destination name.\r\n",57);
    LOG_SEND(      "350 File or dirctory exist, ready for destination name.\r\n",57);

}

//*****************************************************************************
//*
//*     CmdRnto
//*
//*****************************************************************************
//  Handle a RNTO command.    RNTO <path>
//  pData       : Ponter to the connection data
//  pLine       : Ponter to the command line
static void CmdRnto(FtpData *pData,char *pLine)
{
char        cPath[256];
TcpHandle   hClient;
int         iShare;
int         iRet;



    hClient = pData->hConnection;

    if(!pData->cLoginDone)                              // username allready set
        {
        TcpPut(hClient,"530 Not logged in.\r\n",20);
        LOG_SEND(      "530 Not logged in.\r\n",20);
        return;
        }

    if(!pData->cMovePath[0])
        {
        TcpPut(hClient,"553 Requested action not taken.\r\n",33);
        LOG_SEND(      "553 Requested action not taken.\r\n",33);
        pData->cMovePath[0]=0;
        return;
        }

       iShare = CreatePathName(pData,pLine,cPath,sizeof(cPath));
    if(iShare<0)
        {
        TcpPut(hClient,"553 Requested action not taken.\r\n",33);
        LOG_SEND(      "553 Requested action not taken.\r\n",33);
        pData->cMovePath[0]=0;
        return;
        }

    if(pData->aShares[iShare].cReadOnly)                // is it a read oly path
        {
        TcpPut(hClient,"550 Requested action not taken.\r\n",33);
        LOG_SEND(      "550 Requested action not taken.\r\n",33);
        return;
        }

       iRet = SysRename(pData->cMovePath,cPath);        // rename file
    if(iRet==SYS_ERR_FILE_NOT_FOUND || iRet==SYS_ERR_PATH_NOT_FOUND)
        {
        TcpPut(hClient,"553 Requested action not taken.\r\n",33);
        LOG_SEND(      "553 Requested action not taken.\r\n",33);
        }
    else if(iRet!=SYS_ERR_NONE)
        {
        TcpPut(hClient,"550 Requested action not taken.\r\n",33);
        LOG_SEND(      "550 Requested action not taken.\r\n",33);
        }
    else{
        TcpPut(hClient,"250 Requested file action okay, completed.\r\n",44);
        LOG_SEND(      "250 Requested file action okay, completed.\r\n",44);
        }

    pData->cMovePath[0]=0;

}

//*****************************************************************************
//*
//*     CmdDele
//*
//*****************************************************************************
//  Handle a DELE command.    DELE <path>
//  pData       : Ponter to the connection data
//  pLine       : Ponter to the command line
static void CmdDele(FtpData *pData,char *pLine)
{
char        cPath[256];
TcpHandle   hClient;
int         iShare;
int         iRet;


    hClient = pData->hConnection;

    if(!pData->cLoginDone)                              // username allready set
        {
        TcpPut(hClient,"530 Not logged in.\r\n",20);
        LOG_SEND(      "530 Not logged in.\r\n",20);
        return;
        }

       iShare = CreatePathName(pData,pLine,cPath,sizeof(cPath),1);
    if(iShare<0)
        {
        TcpPut(hClient,"553 Requested action not taken.\r\n",33);
        LOG_SEND(      "553 Requested action not taken.\r\n",33);
        return;
        }

    if(pData->aShares[iShare].cReadOnly)                // is it a read oly path
        {
        TcpPut(hClient,"550 Requested action not taken.\r\n",33);
        LOG_SEND(      "550 Requested action not taken.\r\n",33);
        return;
        }

       iRet = SysDelete(cPath);                         // delete file
    if(iRet==SYS_ERR_FILE_NOT_FOUND || iRet==SYS_ERR_PATH_NOT_FOUND)
        {
        TcpPut(hClient,"553 Requested action not taken.\r\n",33);
        LOG_SEND(      "553 Requested action not taken.\r\n",33);
        }
    else if(iRet!=SYS_ERR_NONE)
        {
        TcpPut(hClient,"550 Requested action not taken.\r\n",33);
        LOG_SEND(      "550 Requested action not taken.\r\n",33);
        }
    else{
        TcpPut(hClient,"250 Requested file action okay, completed.\r\n",44);
        LOG_SEND(      "250 Requested file action okay, completed.\r\n",44);
        }


}

//*****************************************************************************
//*
//*     CmdRmd
//*
//*****************************************************************************
//  Handle a RMD command.    RMD <path>
//  pData       : Ponter to the connection data
//  pLine       : Ponter to the command line
static void CmdRmd(FtpData *pData,char *pLine)
{
char        cPath[256];
TcpHandle   hClient;
int         iShare;
int         iRet;


    hClient = pData->hConnection;

    if(!pData->cLoginDone)                              // username allready set
        {
        TcpPut(hClient,"530 Not logged in.\r\n",20);
        LOG_SEND(      "530 Not logged in.\r\n",20);
        return;
        }

       iShare = CreatePathName(pData,pLine,cPath,sizeof(cPath),1);
    if(iShare<0 || pData->aShares[iShare].iPathLen>=pData->iPathLen)
        {
        TcpPut(hClient,"553 Requested action not taken.\r\n",33);
        LOG_SEND(      "553 Requested action not taken.\r\n",33);
        return;
        }

    if(pData->aShares[iShare].cReadOnly)                // is it a read only path
        {
        TcpPut(hClient,"550 Requested action not taken.\r\n",33);
        LOG_SEND(      "550 Requested action not taken.\r\n",33);
        return;
        }

       iRet = SysRmdir(cPath);                          // remove directory
    if(iRet==SYS_ERR_FILE_NOT_FOUND || iRet==SYS_ERR_PATH_NOT_FOUND)
        {
        TcpPut(hClient,"553 Requested action not taken.\r\n",33);
        LOG_SEND(      "553 Requested action not taken.\r\n",33);
        }
    else if(iRet!=SYS_ERR_NONE)
        {
        TcpPut(hClient,"550 Requested action not taken.\r\n",33);
        LOG_SEND(      "550 Requested action not taken.\r\n",33);
        }
    else{
        TcpPut(hClient,"250 Requested file action okay, completed.\r\n",44);
        LOG_SEND(      "250 Requested file action okay, completed.\r\n",44);
        }


}

//*****************************************************************************
//*
//*     CmdMkd
//*
//*****************************************************************************
//  Handle a MKD command.    MKD <path>
//  pData       : Ponter to the connection data
//  pLine       : Ponter to the command line
static void CmdMkd(FtpData *pData,char *pLine)
{
char        cPath[256];
char        cTemp[256];
TcpHandle   hClient;
int         iShare;
int         iRet;
int         iLen;


    hClient = pData->hConnection;

    if(!pData->cLoginDone)                              // username allready set
        {
        TcpPut(hClient,"530 Not logged in.\r\n",20);
        LOG_SEND(      "530 Not logged in.\r\n",20);
        return;
        }

       iShare = CreatePathName(pData,pLine,cPath,sizeof(cPath));
    if(iShare<0 || pData->iPathRoot || pData->aShares[iShare].iPathLen>=pData->iPathLen)
        {
        TcpPut(hClient,"553 Requested action not taken.\r\n",33);
        LOG_SEND(      "553 Requested action not taken.\r\n",33);
        return;
        }

    if(pData->aShares[iShare].cReadOnly)                // is the path a read only path
        {
        TcpPut(hClient,"550 Requested action not taken.\r\n",33);
        LOG_SEND(      "550 Requested action not taken.\r\n",33);
        return;
        }

       iRet = SysMkdir(cPath);                          // create directory
    if(iRet==SYS_ERR_FILE_NOT_FOUND || iRet==SYS_ERR_PATH_NOT_FOUND)
        {
        TcpPut(hClient,"553 Requested action not taken.\r\n",33);
        LOG_SEND(      "553 Requested action not taken.\r\n",33);
        }
    else if(iRet!=SYS_ERR_NONE)
        {
        TcpPut(hClient,"550 Requested action not taken.\r\n",33);
        LOG_SEND(      "550 Requested action not taken.\r\n",33);
        }
    else{
        if(pData->aShares[iShare].iNameLen)
            {
            iLen = sprintf(cTemp,"257 \"/%s/%s\" directory created\r\n",pData->aShares[iShare].cName,cPath+pData->aShares[iShare].iPathLen);
            }
        else{
            iLen = sprintf(cTemp,"257 \"/%s\" directory created\r\n",cPath+pData->aShares[iShare].iPathLen);
            }

        TcpPut(hClient,cTemp,iLen);
        LOG_SEND(      cTemp,iLen);
        }


}

//*****************************************************************************
//*
//*     CmdMdtm
//*
//*****************************************************************************
//  Handle a MDTM command.    MDTM <path>
//  pData       : Ponter to the connection data
//  pLine       : Ponter to the command line
static void CmdMdtm(FtpData *pData,char *pLine)
{
SysTimeStruct   sTime;
SysFindStruct   sFind;
SysFind         hFind;
SysFile         hFile;
char            cPath[256];
char            cTemp[256];
TcpHandle       hClient;
int             iShare;
int             iLen;
int             iPos;
int             iRet;



    hClient = pData->hConnection;


    if(!pData->cLoginDone)                              // username allready set
        {
        TcpPut(hClient,"530 Not logged in.\r\n",20);
        LOG_SEND(      "530 Not logged in.\r\n",20);
        return;
        }


    for(iPos=0;;iPos++)                                 // seach for second parameter
        {
        if(pLine[iPos]=='\0')break;
        if(pLine[iPos]=='\r')break;
        if(pLine[iPos]=='\n')break;
        if(pLine[iPos]!=' ' )continue;

        if(iPos!=14 && iPos!=8)
            {
            TcpPut(hClient,"501 Syntax error in parameters or arguments.\r\n",46);
            LOG_SEND(      "501 Syntax error in parameters or arguments.\r\n",46);
            return;
            }

        iPos   += 1;
        iShare  = CreatePathName(pData,pLine+iPos,cPath,sizeof(cPath),1);

        if(iShare<0 || pData->iPathRoot || (pData->iPathLen>0 && cPath[pData->iPathLen-1]=='/'))
            {
            TcpPut(hClient,"553 Requested action not taken.\r\n",33);
            LOG_SEND(      "553 Requested action not taken.\r\n",33);
            return;
            }

        if(iPos>=14)                                    // with time
            {
            pLine[14] = 0;  sTime.ucSecond=atoi(pLine+12);
            pLine[12] = 0;  sTime.ucMinute=atoi(pLine+10);
            pLine[10] = 0;  sTime.ucHour  =atoi(pLine+ 8);
            }
        else{
            sTime.usYear  = 0;
            sTime.ucMonth = 0;
            sTime.ucDay   = 0;
            }

        pLine[8] = 0;   sTime.ucDay  =atoi(pLine+6);
        pLine[6] = 0;   sTime.ucMonth=atoi(pLine+4);
        pLine[4] = 0;   sTime.usYear =atoi(pLine+0);
                                                        // check date
        if(sTime.usYear  <1900 || sTime.usYear >2200 ||
           sTime.ucMonth <   1 || sTime.ucMonth>  12 ||
           sTime.ucDay   <   1 || sTime.ucDay  >  31 ||
           sTime.ucHour  >  59 ||
           sTime.ucMinute>  59 ||
           sTime.ucSecond>  59)
            {
            TcpPut(hClient,"501 Syntax error in parameters or arguments.\r\n",46);
            LOG_SEND(      "501 Syntax error in parameters or arguments.\r\n",46);
            return;
            }


        sTime.ucIsDst     = 0;
        sFind.dwTimeWrite = SysTimeFromStruct(&sTime,0);


        if(sFind.dwTimeWrite==0xFFFFFFFF || pData->aShares[iShare].cReadOnly)
            {
            TcpPut(hClient,"550 Requested action not taken.\r\n",33);
            LOG_SEND(      "550 Requested action not taken.\r\n",33);
            return;
            }

            hFile = SysOpen(cPath,SYS_WRITE);           // open file
        if(!hFile)
            {
            TcpPut(hClient,"550 Requested action not taken.\r\n",33);
            LOG_SEND(      "550 Requested action not taken.\r\n",33);
            return;
            }
                                                        // set write time
           iRet = SysSetFileTime(hFile,sFind.dwTimeWrite,0,0);
        if(iRet<0)
            {
            TcpPut(hClient,"550 Requested action not taken.\r\n",33);
            LOG_SEND(      "550 Requested action not taken.\r\n",33);
            return;
            }

        TcpPut(hClient,"213 Command successful.\r\n",25);
        LOG_SEND(      "213 Command successful.\r\n",25);
        return;
        }



       iShare = CreatePathName(pData,pLine,cPath,sizeof(cPath));
    if(iShare<0 || pData->iPathRoot || (pData->iPathLen>0 && cPath[pData->iPathLen-1]=='/'))
        {
        TcpPut(hClient,"553 Requested action not taken.\r\n",33);
        LOG_SEND(      "553 Requested action not taken.\r\n",33);
        return;
        }


        hFind = SysFindFirst(cPath,&sFind);             // get file info
    if(!hFind)
        {
        TcpPut(hClient,"550 Requested action not taken.\r\n",33);
        LOG_SEND(      "550 Requested action not taken.\r\n",33);
        return;
        }


    SysFindClose(hFind);
    SysTimeToStruct(&sTime,sFind.dwTimeWrite,0);


    iLen = sprintf(cTemp,"213 %04u%02u%02u%02u%02u%02u\r\n",sTime.usYear,sTime.ucMonth,sTime.ucDay,sTime.ucHour,sTime.ucMinute,sTime.ucSecond);
    TcpPut(hClient,cTemp,iLen);
    LOG_SEND(      cTemp,iLen);


}

//*****************************************************************************
//*
//*     CmdSize
//*
//*****************************************************************************
//  Handle a SIZE command.    SIZE <path>
//  pData       : Ponter to the connection data
//  pLine       : Ponter to the command line
static void CmdSize(FtpData *pData,char *pLine)
{
SysFindStruct   sFind;
SysFind         hFind;
char            cPath[256];
char            cTemp[256];
TcpHandle       hClient;
int             iShare;
int             iLen;




    hClient = pData->hConnection;


    if(!pData->cLoginDone)                              // username allready set
        {
        TcpPut(hClient,"530 Not logged in.\r\n",20);
        LOG_SEND(      "530 Not logged in.\r\n",20);
        return;
        }


       iShare = CreatePathName(pData,pLine,cPath,sizeof(cPath));
    if(iShare<0 || pData->iPathRoot || (pData->iPathLen>0 && cPath[pData->iPathLen-1]=='/'))
        {
        TcpPut(hClient,"553 Requested action not taken.\r\n",33);
        LOG_SEND(      "553 Requested action not taken.\r\n",33);
        return;
        }


        hFind = SysFindFirst(cPath,&sFind);             // get file info
    if(!hFind)
        {
        TcpPut(hClient,"550 Requested action not taken.\r\n",33);
        LOG_SEND(      "550 Requested action not taken.\r\n",33);
        return;
        }


    SysFindClose(hFind);


    iLen = sprintf(cTemp,"213 %u\r\n",sFind.dwSize);
    TcpPut(hClient,cTemp,iLen);
    LOG_SEND(      cTemp,iLen);


}

//*****************************************************************************
//*
//*     CmdList
//*
//*****************************************************************************
//  Handle a LIST command.    LIST <path>
//  pData       : Ponter to the connection data
//  pLine       : Ponter to the command line
static void CmdList(FtpData *pData,char *pLine)
{
void               *pParam;
char               *pName;
char                cPath[256];
char                cData[352];
char                cTemp[ 16];
char                cTime[ 16];
SysUtc              uTime;
SysFind             hFind;
SysFindStruct       sData;
SysTimeStruct       sTime;
FtpFileCheckerCB    pChecker;
TcpHandle           hClient;
int                 iShare;
int                 iOnly;
int                 iFlag;
int                 iRet;
int                 iLen;
int                 iPos;
int                 iNum;


    hClient = pData->hConnection;

    if(!pData->cLoginDone)                              // username allready set
        {
        TcpPut(hClient,"530 Not logged in.\r\n",20);
        LOG_SEND(      "530 Not logged in.\r\n",20);
        FtpCloseListenPort(pData);
        return;
        }


	if(*pLine=='-')										// skip attributes
		{
		for(pLine++;;pLine++)
			{
			if(*pLine== 0  )break;
			if(*pLine==' ' )break;
			if(*pLine=='\r')break;
			if(*pLine=='\n')break;
			}

		for(;*pLine==' ';pLine++);
		}


       iShare  = CreatePathName(pData,pLine,cPath,sizeof(cPath)-2,2);
    if(iShare<0)
        {
        if(!pData->iPathRoot)
            {
            TcpPut(hClient,"553 Requested action not taken.\r\n",33);
            LOG_SEND(      "553 Requested action not taken.\r\n",33);
            FtpCloseListenPort(pData);
            return;
            }

        iLen = 0;
        }
    else{
        iLen = pData->iPathLen;                         // are we in a real folder

        if(iLen>0 && cPath[iLen-1]=='/')
            {
            iLen--;

            if(iLen>3 || cPath[1]!=':')
                {
                cPath[iLen] = 0;
                }
            }

           iRet = SysGetAttributes(cPath);
        if(iRet<0 || !(iRet&SYS_DIR))
            {
            TcpPut(hClient,"553 Requested action not taken.\r\n",33);
            LOG_SEND(      "553 Requested action not taken.\r\n",33);
            FtpCloseListenPort(pData);
            return;
            }

                                                        // exists a file checker callback
        if(FTP_CHECKER(pData->aShares[iShare].pCbChecker) && iLen>pData->aShares[iShare].iPathLen)
            {
            if(CheckPath(pData,cPath,iLen,pData->aShares[iShare].iPathLen,iShare))
                {
                TcpPut(hClient,"553 Requested action not taken.\r\n",33);
                LOG_SEND(      "553 Requested action not taken.\r\n",33);
                FtpCloseListenPort(pData);
                return;
                }
            }
        }


    TcpPut(hClient,"150 Opening ASCII mode data connection for file list.\r\n",55);
    LOG_SEND(      "150 Opening ASCII mode data connection for file list.\r\n",55);


        pData->hDataConn = FtpOpenDataPort(pData);      // get data connection
    if(!pData->hDataConn)
        {
        TcpPut(hClient,"425 Cant open data connection.\r\n",33);
        LOG_SEND(      "425 Cant open data connection.\r\n",33);
        return;
        }


    uTime = SysTimeGet();

    if(iShare<0 || pData->iPathRoot)                    // return the share list
        {
        pName = ".";
        iOnly = 0;

        for(iPos=0;iPos<=pData->iShareCount;iPos++)
            {
            SysWatchdogTrigger(pData->hWatchdog);

            if(iPos==pData->iShareRoot)continue;

            iNum = sprintf(cData,"dr%c-r%c----   0 ftp ftp           1 Jan  1  2000 %s\r\n",(iOnly)? '-':'w',(iOnly)? '-':'w',pName);

            if(TcpPut(pData->hDataConn,cData,iNum)!=iNum)
                {
                TcpClose(pData->hDataConn);
                pData->hDataConn=0;
                TcpPut(hClient,"426 Connection closed; transfer aborted.\r\n",42);
                LOG_SEND(      "426 Connection closed; transfer aborted.\r\n",42);
                return;
                }

            pName = pData->aShares[iPos].cName;
            iOnly = pData->aShares[iPos].cReadOnly;
            }
        }


    if(iShare>=0)                                       // find files in the directory
        {
        cPath[iLen  ] = '/';
        cPath[iLen+1] = '*';
        cPath[iLen+2] = 0;

        hFind         = SysFindFirst(cPath,&sData);
        iFlag         = (pData->aShares[iShare].cReadOnly)? ~SYS_WRITE:-1;
        pParam        =  pData->aShares[iShare].pCbParam;
        pChecker      =  pData->aShares[iShare].pCbChecker;
        cPath[iLen]   =  0;
        }
    else{
        hFind         =  0;
        iFlag         = -1;
        }


    while(hFind)                                        // output the file list
        {
        SysThreadCheck();
        SysWatchdogTrigger(pData->hWatchdog);

        if(sData.cName[0]=='.' && pData->iPathRoot)     // hide '..' at root path
            {
            if(sData.cName[1]==0 || (sData.cName[1]=='.' && sData.cName[2]==0))
                {
                if(SysFindNext(hFind,&sData)<0)
                    {
                    SysFindClose(hFind);
                    hFind=0;
                    }

                continue;
                }
            }


        if(pChecker)                                    // exists a file checker callback
            {
            iRet = pChecker(pData,cPath,sData.cName,pParam,iShare);

            if(iRet>=3)                                 // hide file
                {
                if(SysFindNext(hFind,&sData)<0)
                    {
                    SysFindClose(hFind);
                    hFind=0;
                    }

                continue;
                }

            if(iRet>=1)                                 // is file read only
                {
                sData.dwAttr &= ~SYS_WRITE;
                }
            }



        SysTimeToStruct(&sTime,sData.dwTimeWrite,0);
        sData.dwAttr &= iFlag;

                                                        // filetime in future or older than 6 mounths
        if(sData.dwTimeWrite>uTime || uTime-sData.dwTimeWrite>60*60*24*182)
            {
            sprintf(cTime," %04u",sTime.usYear);
            }
        else{
            sprintf(cTime,"%02u:%02u",sTime.ucHour,sTime.ucMinute);
            }


        iNum = sprintf(cData,"%c%c%c%c%c%c%c%c--   0 ftp ftp %11i %s %02u %s %s\r\n",
                        (sData.dwAttr&SYS_DIR  )? 'd':'-',
                        (sData.dwAttr&SYS_READ )? 'r':'-',
                        (sData.dwAttr&SYS_WRITE)? 'w':'-',
                        (sData.dwAttr&SYS_EXEC )? 'x':'-',
                        (sData.dwAttr&SYS_READ )? 'r':'-',
                        (sData.dwAttr&SYS_WRITE)? 'w':'-',
                        (sData.dwAttr&SYS_EXEC )? 'x':'-',
                        (sData.dwAttr&SYS_READ )? 'r':'-',
                         sData.dwSize,
                         pMonthNames[sTime.ucMonth&0x0F],
                         sTime.ucDay,
                         cTime,
                         sData.cName);


        if(TcpPut(pData->hDataConn,cData,iNum)!=iNum)   // transmitt entry
            {
            TcpClose(pData->hDataConn);
            pData->hDataConn=0;
            SysFindClose(hFind);
            TcpPut(hClient,"426 Connection closed; transfer aborted.\r\n",42);
            LOG_SEND(      "426 Connection closed; transfer aborted.\r\n",42);
            return;
            }

        LOG_LIST(cData,iNum);

        if(SysFindNext(hFind,&sData)<0)
            {
            SysFindClose(hFind);
            hFind=0;
            }

           iLen = TcpGetCount(hClient);
        if(iLen==6)                                     // check for abort
            {
            TcpGet(hClient,cTemp,6);
            if(!memcmp(cTemp,"ABOR\r\n",6))
                {
                TcpClose(pData->hDataConn);
                pData->hDataConn=0;
                SysFindClose(hFind);
                TcpPut(hClient,"426 The service request terminated abnormally\r\n226 The abort command was successfully processed\r\n",97);
                LOG_SEND(      "426 The service request terminated abnormally\r\n226 The abort command was successfully processed\r\n",97);
                return;
                }
            }
        }


    TcpClose(pData->hDataConn);
    pData->hDataConn=0;


    TcpPut(hClient,"226 Transfer complete.\r\n",24);
    LOG_SEND(      "226 Transfer complete.\r\n",24);


}


//*****************************************************************************
//*
//*     CmdRetr
//*
//*****************************************************************************
//  Handle a RETR command.    RETR <path>
//  pData       : Ponter to the connection data
//  pLine       : Ponter to the command line
static void CmdRetr(FtpData *pData,char *pLine)
{
char            cPath[512];
SysFile         hFile;
TcpHandle       hClient;
int             iShare;
int             iRet;
int             iLen;


    hClient = pData->hConnection;

    if(!pData->cLoginDone)                              // username allready set
        {
        TcpPut(hClient,"530 Not logged in.\r\n",20);
        LOG_SEND(      "530 Not logged in.\r\n",20);
        FtpCloseListenPort(pData);
        return;
        }

       iShare = CreatePathName(pData,pLine,cPath,sizeof(cPath)-2);
    if(iShare<0)
        {
        TcpPut(hClient,"553 Requested action not taken.\r\n",33);
        LOG_SEND(      "553 Requested action not taken.\r\n",33);
        FtpCloseListenPort(pData);
        return;
        }

    iLen  = strlen(cPath);

    if(iLen>0 && cPath[iLen-1]=='/')
        {
        iLen--;

        if(iLen>3 || cPath[1]!=':')
            {
            cPath[iLen] = 0;
            }
        }


       iRet = SysGetAttributes(cPath);                  // get file attributes
    if(iRet<0 || (iRet&SYS_DIR))
        {
        TcpPut(hClient,"553 Requested action not taken.\r\n",33);
        LOG_SEND(      "553 Requested action not taken.\r\n",33);
        FtpCloseListenPort(pData);
        return;
        }


    TcpPut(hClient,"150 Opening BINARY mode data connection.\r\n",42);
    LOG_SEND(      "150 Opening BINARY mode data connection.\r\n",42);


        pData->hDataConn = FtpOpenDataPort(pData);
    if(!pData->hDataConn)                               // get data connection
        {
        TcpPut(hClient,"425 Cant open data connection.\r\n",33);
        LOG_SEND(      "425 Cant open data connection.\r\n",33);
        return;
        }


        hFile = SysOpen(cPath,SYS_READ);
    if(!hFile)
        {
        TcpPut(hClient,"553 Requested action not taken.\r\n",33);
        LOG_SEND(      "553 Requested action not taken.\r\n",33);
        TcpClose(pData->hDataConn);
        pData->hDataConn=0;
        return;
        }


    for(;;)                                             // output the file
        {
        SysThreadCheck();
        SysWatchdogTrigger(pData->hWatchdog);

        iLen = SysRead(hFile,cPath,sizeof(cPath));

        if(iLen==0)break;
        if(iLen<=0 || TcpPut(pData->hDataConn,cPath,iLen)!=iLen)
            {
            TcpClose(pData->hDataConn);
            pData->hDataConn=0;
            SysClose(hFile);
            TcpPut(hClient,"426 Connection closed; transfer aborted.\r\n",42);
            LOG_SEND(      "426 Connection closed; transfer aborted.\r\n",42);
            return;
            }

           iLen = TcpGetCount(hClient);
        if(iLen==6)                                     // check for abort
            {
            TcpGet(hClient,cPath,6);
            if(!memcmp(cPath,"ABOR\r\n",6))
                {
                TcpPut(hClient,"426 The service request terminated abnormally\r\n226 The abort command was successfully processed\r\n",97);
                LOG_SEND(      "426 The service request terminated abnormally\r\n226 The abort command was successfully processed\r\n",97);
                TcpClose(pData->hDataConn);
                pData->hDataConn=0;
                SysClose(hFile);
                return;
                }
            }
        }


    TcpClose(pData->hDataConn);
    SysClose(hFile);
    pData->hDataConn=0;


    TcpPut(hClient,"226 Transfer complete.\r\n",24);
    LOG_SEND(      "226 Transfer complete.\r\n",24);


}

//*****************************************************************************
//*
//*     CmdStor
//*
//*****************************************************************************
//  Handle a STOR command.    STOR <path>
//  pData       : Ponter to the connection data
//  pLine       : Ponter to the command line
static void CmdStor(FtpData *pData,char *pLine)
{
char            cPath[512];
SysFile         hFile;
TcpHandle       hClient;
int             iShare;
int             iLen;


    hClient = pData->hConnection;

    if(!pData->cLoginDone)                              // username allready set
        {
        TcpPut(hClient,"530 Not logged in.\r\n",20);
        LOG_SEND(      "530 Not logged in.\r\n",20);
        FtpCloseListenPort(pData);
        return;
        }

       iShare  = CreatePathName(pData,pLine,cPath,sizeof(cPath)-2,1);
    if(iShare<0)
        {
        TcpPut(hClient,"553 Requested action not taken.\r\n",33);
        LOG_SEND(      "553 Requested action not taken.\r\n",33);
        FtpCloseListenPort(pData);
        return;
        }

    iLen  = strlen(cPath);

    if(iLen>0 && cPath[iLen-1]=='/')
        {
        iLen--;

        if(iLen>3 || cPath[1]!=':')
            {
            cPath[iLen] = 0;
            }
        }


    if(pData->aShares[iShare].cReadOnly)
        {
        TcpPut(hClient,"553 Requested action not taken.\r\n",33);
        LOG_SEND(      "553 Requested action not taken.\r\n",33);
        FtpCloseListenPort(pData);
        return;
        }


    TcpPut(hClient,"150 Opening BINARY mode data connection.\r\n",42);
    LOG_SEND(      "150 Opening BINARY mode data connection.\r\n",42);


        pData->hDataConn = FtpOpenDataPort(pData);
    if(!pData->hDataConn)                               // get data connection
        {
        TcpPut(hClient,"425 Cant open data connection.\r\n",33);
        LOG_SEND(      "425 Cant open data connection.\r\n",33);
        return;
        }


              hFile = SysOpen(cPath,SYS_WRITE|SYS_TRUNC|SYS_CREATE);
    if(!hFile)hFile = SysOpen(cPath,SYS_WRITE|SYS_TRUNC);
    if(!hFile)
        {
        TcpPut(hClient,"553 Requested action not taken.\r\n",33);
        LOG_SEND(      "553 Requested action not taken.\r\n",33);
        TcpClose(pData->hDataConn);
        pData->hDataConn=0;
        return;
        }



    for(;;)                                             // read the file
        {
        SysThreadCheck();
        SysWatchdogTrigger(pData->hWatchdog);


           iLen = TcpGet(pData->hDataConn,cPath,sizeof(cPath));
        if(iLen<=0)break;

        if(SysWrite(hFile,cPath,iLen)!=iLen)
            {
            TcpPut(hClient,"426 Connection closed; transfer aborted.\r\n",42);
            LOG_SEND(      "426 Connection closed; transfer aborted.\r\n",42);
            TcpClose(pData->hDataConn);
            pData->hDataConn=0;
            SysClose(hFile);
            return;
            }

           iLen = TcpGetCount(hClient);
        if(iLen==6)                                     // check for abort
            {
            TcpGet(hClient,cPath,6);
            if(!memcmp(cPath,"ABOR\r\n",6))
                {
                TcpPut(hClient,"426 The service request terminated abnormally\r\n226 The abort command was successfully processed\r\n",97);
                LOG_SEND(      "426 The service request terminated abnormally\r\n226 The abort command was successfully processed\r\n",97);
                TcpClose(pData->hDataConn);
                pData->hDataConn=0;
                SysClose(hFile);
                return;
                }
            }
        }


    TcpClose(pData->hDataConn);
    SysClose(hFile);
    pData->hDataConn=0;


    TcpPut(hClient,"226 Transfer complete.\r\n",24);
    LOG_SEND(      "226 Transfer complete.\r\n",24);


}

//*****************************************************************************
//*
//*     CmdSite
//*
//*****************************************************************************
//  Handle a SIDE command.    SIDE <subcmd> <path>
//  pData       : Ponter to the connection data
//  pLine       : Ponter to the command line
static void CmdSite(FtpData *pData,char *pLine)
{
char        cPath[256];
TcpHandle   hClient;
int         iShare;
int         iBits;
int         iAttr;
int         iRet;


    hClient = pData->hConnection;

    if(!pData->cLoginDone)                              // username allready set
        {
        TcpPut(hClient,"530 Not logged in.\r\n",20);
        LOG_SEND(      "530 Not logged in.\r\n",20);
        return;
        }

    if(memcmp(pLine,"HELP",5)==0)                       // list all side commands
        {
        TcpPut(hClient,"HELP : display this message\r\nchmod <attrib> <file> : change file premission\r\n200 Command okay.\r\n",96);
        LOG_SEND(      "HELP : display this message\r\nchmod <attrib> <file> : change file premission\r\n200 Command okay.\r\n",96);
        return;
        }

    if(memcmp(pLine,"chmod ",6) || pLine[6]<'0' || pLine[6]>'7' || pLine[7]<'0' || pLine[7]>'7' || pLine[8]<'0' || pLine[8]>'7' || pLine[9]!=' ')
        {
        TcpPut(hClient,"502 Command not implemented.\r\n",30);
        LOG_SEND(      "502 Command not implemented.\r\n",30);
        return;
        }

    iBits = pLine[6]-'0';                               // get attributes

               iAttr = 0;
    if(iBits&4)iAttr|=SYS_READ;
    if(iBits&2)iAttr|=SYS_WRITE;
    if(iBits&1)iAttr|=SYS_EXEC;

    pLine += 10;


       iShare = CreatePathName(pData,pLine,cPath,sizeof(cPath));
    if(iShare<0)
        {
        TcpPut(hClient,"553 Requested action not taken.\r\n",33);
        LOG_SEND(      "553 Requested action not taken.\r\n",33);
        return;
        }

    if(pData->aShares[iShare].cReadOnly)                // is it a read oly path
        {
        TcpPut(hClient,"550 Requested action not taken.\r\n",33);
        LOG_SEND(      "550 Requested action not taken.\r\n",33);
        return;
        }



       iRet = SysSetAttributes(cPath,iAttr);            // delete file
    if(iRet==SYS_ERR_FILE_NOT_FOUND || iRet==SYS_ERR_PATH_NOT_FOUND)
        {
        TcpPut(hClient,"553 Requested action not taken.\r\n",33);
        LOG_SEND(      "553 Requested action not taken.\r\n",33);
        }
    else if(iRet!=SYS_ERR_NONE)
        {
        TcpPut(hClient,"550 Requested action not taken.\r\n",33);
        LOG_SEND(      "550 Requested action not taken.\r\n",33);
        }
    else{
        TcpPut(hClient,"250 Requested file action okay, completed.\r\n",44);
        LOG_SEND(      "250 Requested file action okay, completed.\r\n",44);
        }


}

//*****************************************************************************
//*
//*     FtpConnectionProc
//*
//*****************************************************************************
//  Function which receives the command from the client
//  pParam      : pointer to the config data of the server
static void _cdecl FtpConnectionProc(void *pParam)
{
FtpData        *pData=(FtpData*)pParam;
FtpServerData  *pServer;
char           *pLine;
TcpHandle       hClient;
char            cTemp[256];
int             iCmd;
int             iLen;




    pData->hWatchdog    = SysWatchdogRegister(FTP_WDOG_TIME,"FtpConnection");
    pServer             = pData->pServer;
    hClient             = pData->hConnection;
    iLen                = sprintf(cTemp,"220 ("FTP_VERSION")\r\n");


    TcpPut(hClient,cTemp,iLen);
    LOG_SEND(      cTemp,iLen);

    while(!pServer->iShutdown)
        {
        SysThreadCheck();
        SysWatchdogTrigger(pData->hWatchdog);


           iLen= FtpGetLine(pData,&pLine,pServer->uTimeout);
        if(iLen<3)break;


        LOG_RECV(pLine,iLen);


        iCmd  = *(int*)pLine;
        iCmd &= 0xDFDFDFDF;


        if(iCmd==*(int*)"PWD")                          // get working directory
            {
            SysThreadCheck();
            CmdPwd(pData);
            continue;
            }


        if(iCmd==*(int*)"CWD" && pLine[3]==' ')         // change working directory
            {
            SysThreadCheck();
            CmdCwd(pData,pLine+4);
            continue;
            }


        if(iCmd==*(int*)"MKD" && pLine[3]==' ')         // create a directory
            {
            SysThreadCheck();
            CmdMkd(pData,pLine+4);
            continue;
            }


        if(iCmd==*(int*)"RMD" && pLine[3]==' ')         // delete a directory
            {
            SysThreadCheck();
            CmdRmd(pData,pLine+4);
            continue;
            }


        if(iCmd==*(int*)"LIST" && pLine[4]==' ')        // list all files
            {
            SysThreadCheck();
            CmdList(pData,pLine+5);
            continue;
            }


        if(iCmd==*(int*)"LIST" && pLine[4]==0)          // list all files of current directory
            {
            SysThreadCheck();
            CmdList(pData,pLine+4);
            continue;
            }


        if(iCmd==*(int*)"RETR" && pLine[4]==' ')        // read a file
            {
            SysThreadCheck();
            CmdRetr(pData,pLine+5);
            continue;
            }


        if(iCmd==*(int*)"STOR" && pLine[4]==' ')        // save a file
            {
            SysThreadCheck();
            CmdStor(pData,pLine+5);
            continue;
            }


        if(iCmd==*(int*)"DELE" && pLine[4]==' ')        // delete a file
            {
            SysThreadCheck();
            CmdDele(pData,pLine+5);
            continue;
            }


        if(iCmd==*(int*)"PASV" && pLine[4]==0)          // activate passive connection
            {
            SysThreadCheck();
            CmdPasv(pData);
            continue;
            }


        if(iCmd==*(int*)"EPSV" && pLine[4]==0)          // activate passive connection (extended)
            {
            SysThreadCheck();
            CmdEpsv(pData);
            continue;
            }


        if(iCmd==*(int*)"CDUP" && pLine[4]==0)          // path up in working directory
            {
            SysThreadCheck();
            CmdCdup(pData);
            continue;
            }


        if(iCmd==*(int*)"RNFR" && pLine[4]==' ')        // set name for rename file
            {
            SysThreadCheck();
            CmdRnfr(pData,pLine+5);
            continue;
            }


        if(iCmd==*(int*)"RNTO" && pLine[4]==' ')        // set rename file
            {
            SysThreadCheck();
            CmdRnto(pData,pLine+5);
            continue;
            }


        if(iCmd==*(int*)"USER" && pLine[4]==' ')        // set user name
            {
            SysThreadCheck();
            CmdUser(pData,pLine+5);
            continue;
            }


        if(iCmd==*(int*)"PASS" && pLine[4]==' ')        // password input
            {
            SysThreadCheck();
            CmdPass(pData,pLine+5);
            continue;
            }


        if(iCmd==*(int*)"MODE" && pLine[4]==' ')        // mode selection
            {
            SysThreadCheck();
            CmdMode(pData,pLine+5);
            continue;
            }


        if(iCmd==*(int*)"TYPE" && pLine[4]==' ')        // type selection
            {
            SysThreadCheck();
            CmdType(pData,pLine+5);
            continue;
            }


        if(iCmd==*(int*)"STRU" && pLine[4]==' ')        // file structure for transfer
            {
            SysThreadCheck();
            CmdStru(pData,pLine+5);
            continue;
            }


        if(iCmd==*(int*)"PORT" && pLine[4]==' ')        // set listening port (active mode)
            {
            SysThreadCheck();
            CmdPort(pData,pLine+5);
            continue;
            }


        if(iCmd==*(int*)"EPRT" && pLine[4]==' ')        // set listening port (active mode, extended)
            {
            SysThreadCheck();
            CmdEprt(pData,pLine+5);
            continue;
            }


        if(iCmd==*(int*)"MDTM" && pLine[4]==' ')        // set/get modification time
            {
            SysThreadCheck();
            CmdMdtm(pData,pLine+5);
            continue;
            }


        if(iCmd==*(int*)"SIZE" && pLine[4]==' ')        // get filesize time
            {
            SysThreadCheck();
            CmdSize(pData,pLine+5);
            continue;
            }


        if(iCmd==*(int*)"SITE" && pLine[4]==' ')        // set attrib
            {
            SysThreadCheck();
            CmdSite(pData,pLine+5);
            continue;
            }

        if(iCmd==*(int*)"FEAT" && pLine[4]==0)          // get features
            {
            SysThreadCheck();
            CmdFeat(pData);
            continue;
            }

        if(iCmd==*(int*)"ABOR" && pLine[4]==0)          // abort command
            {
            SysThreadCheck();
            TcpPut(hClient,"225 Data connection open; no transfer in progress.\r\n",52);
            LOG_SEND(      "225 Data connection open; no transfer in progress.\r\n",52);
            continue;
            }

        if(iCmd==*(int*)"SYST" && pLine[4]==0)          // get operating system
            {
            SysThreadCheck();
            TcpPut(hClient,"215 UNIX Type: L8\r\n",19);
            LOG_SEND(      "215 UNIX Type: L8\r\n",19);
            continue;
            }


        if(iCmd==*(int*)"NOOP" && pLine[4]==0)          // no operation
            {
            SysThreadCheck();
            TcpPut(hClient,"200 Command okay.\r\n",19);
            LOG_SEND(      "200 Command okay.\r\n",19);
            continue;
            }

        if(iCmd==*(int*)"QUIT" && pLine[4]==0)          // cloes connection
            {
            SysThreadCheck();
            TcpPut(hClient,"221 Service closing control connection.\r\n",41);
            LOG_SEND(      "221 Service closing control connection.\r\n",41);
            break;
            }

        SysThreadCheck();

        TcpPut(hClient,"502 Command not implemented.\r\n",30);
        LOG_SEND(      "502 Command not implemented.\r\n",30);
        }



//******************** close connection ***************************************



    TcpClose(hClient);

    if(pData->hDataConn)
        {
        TcpClose(pData->hDataConn);
        }

    if(pData->hListenConn)
        {
        TcpClose(pData->hListenConn);
        }

    if(pServer->pCbDisconnect)
        {
        pServer->pCbDisconnect(pData,pData->cConnIp,pData->usConnPort);
        }

    FreeFtpData(pData);

    SysSemaphoreLock(pServer->hSemaphore);
    pServer->iCurConnections--;
    SysSemaphoreUnlock(pServer->hSemaphore);

    SysWatchdogUnregister(pData->hWatchdog);


}

//*****************************************************************************
//*
//*     FtpListenProc
//*
//*****************************************************************************
//  Function which install the server socket and waits for new connections
//  pParam      : pointer to the config data of the server
static void _cdecl FtpListenProc(void *pParam)
{
FtpServerData  *pServer=(FtpServerData*)pParam;
FtpData        *pData;
SysWdog         hWdog;
unsigned short  usPort;
TcpHandle       hClient;
IpAddr          iAddr;





    hWdog = SysWatchdogRegister(FTP_WDOG_TIME,"FtpListen");



    while(!pServer->iShutdown)
        {
        SysThreadCheck();
        SysWatchdogTrigger(hWdog);

            hClient = TcpAcceptWait(pServer->hServer,&iAddr,&usPort,3000);
        if(!hClient)                                        // wait for client connection
            {
            continue;
            }


        SysThreadCheck();

                                                            // check connection count
        if(pServer->iCurConnections>=pServer->iMaxConnections)
            {
            TcpPut  (hClient,"421 Service Unavailable (too much connections open)\r\n",53);
            LOG_SEND(        "421 Service Unavailable (too much connections open)\r\n",53);
            TcpClose(hClient);
            continue;
            }

            pData = AllocFtpData();
        if(!pData)
            {
            TcpPut  (hClient,"421 Service Unavailable (not enouth memory)\r\n",45);
            LOG_SEND(        "421 Service Unavailable (not enouth memory)\r\n",45);
            TcpClose(hClient);
            continue;
            }

        memset(pData,0,sizeof(FtpData));
        pData->iShareRoot       = -1;
        pData->iShareSel        = -1;
        pData->pServer          = pServer;
        pData->usConnPort       = usPort;
        pData->hConnection      = hClient;
        IpToString(&iAddr,pData->cConnIp);

        LOG_CONN(iAddr,usPort,"accept new ftp connection");

        if(pServer->pCbConnect)                             // connect callback
            {
            SysThreadCheck();

            if(!pServer->pCbConnect(pData,pData->cConnIp,pData->usConnPort))
                {
                TcpPut  (hClient,"421 Service Unavailable (not allowed)\r\n",39);
                LOG_SEND(        "421 Service Unavailable (not allowed)\r\n",39);
                TcpClose(hClient);
                FreeFtpData(pData);
                continue;
                }
            }


        SysThreadCheck();

        SysSemaphoreLock(pServer->hSemaphore);              // start connection thread


        if(SysThreadStart(FtpConnectionProc,FTP_STACK_SIZE,pData,pServer->iPriority|SYS_GLOBAL,"FtpConnection"))
            {
            pServer->iCurConnections++;
            }
        else{
            TcpPut  (hClient,"421 Service Unavailable (not allowed)\r\n",39);
            LOG_SEND(        "421 Service Unavailable (not allowed)\r\n",39);
            TcpClose(hClient);
            FreeFtpData(pData);
            }

        SysSemaphoreUnlock(pServer->hSemaphore);
        }



    while(pServer->iCurConnections>0)                       // wait until all connections are closed
        {
        SysWatchdogTrigger(hWdog);
        SysSleep(100);
        }


    SysSemaphoreLock(pServer->hSemaphore);                  // close server socket
    TcpClose(pServer->hServer);
    pServer->hServer = NULL;
    SysSemaphoreUnlock(pServer->hSemaphore);

    SysSemaphoreDelete(pServer->hSemaphore);
    pServer->hSemaphore = 0;
    pServer->iRunning   = 0;

    SysWatchdogUnregister(hWdog);

}


//*****************************************************************************
//*
//*     FtpDaemon
//*
//*****************************************************************************
//  Starts the http server
//  pConfig     : pointer to the config data for the server
//  Returns 0 it daemon was started, or an error code greather zero
//                  FTP_ERR_NONE            0     no error
//                  FTP_ERR_SOCKET          1     socket error
//                  FTP_ERR_ISRUNNING       2     ftp server is allready running
//                  FTP_ERR_WRONGIP         4     the ip address is wrong
//                  FTP_ERR_INIT            5     initialization error
//                  FTP_ERR_WRONGPATH       7     a wrong path was defined
//                  FTP_ERR_NEED_LOGINCB    10    we need a login callback
FTP_API int FtpDaemon(FtpConfigData *pConfig)
{
IpAddr      iAddr;
int         iAttr;
int         iPort;



    if(iServerRunning)                                      // is the server allready running
        {
        return FTP_ERR_ISRUNNING;
        }

    if(pConfig->iNoAnonym && !pConfig->pCbLogin)            // do we need a login callback
        {
        return FTP_ERR_NEED_LOGINCB;
        }

    if(!hFtpHeap)                                           // create heap for server
        {
            hFtpHeap = SysHeapCreate(FTP_MEMORY_MAX);
        if(!hFtpHeap)return FTP_ERR_INIT;
        }

    memset(&sFtpServer,0,sizeof(sFtpServer));


    if(!IpAddress(&iAddr,pConfig->cIpAddress,NULL))         // convert ip-text to address
        {
        return FTP_ERR_WRONGIP;
        }

        sFtpServer.hSemaphore = SysSemaphoreCreate();
    if(!sFtpServer.hSemaphore)                              // create server semaphore
        {
        return FTP_ERR_INIT;
        }


    if(!pConfig->usTcpPort)
        {
        iPort = FTP_PORT;
        }
    else{
        iPort = pConfig->usTcpPort;
        }



        sFtpServer.hServer = TcpOpen(&iAddr,iPort,TCP_SERVER,0);
    if(!sFtpServer.hServer)                                 // check server socket
        {
        SysSemaphoreDelete(sFtpServer.hSemaphore);
        return FTP_ERR_SOCKET;
        }


    if(pConfig->cFilePath[0])                               // check path
        {
        iAttr = SysGetAttributes(pConfig->cFilePath);

        if(iAttr<0 || !(iAttr&SYS_DIR))
            {
            return FTP_ERR_WRONGPATH;
            }
        }

    if(!pConfig->pCbLogin && !pConfig->cFilePath[0])        // need path if no login callback
        {
        return FTP_ERR_WRONGPATH;
        }

    sFtpServer.usTcpPort        = pConfig->usTcpPort;       // is the tcp port for the server
    sFtpServer.iNoAnonym        = pConfig->iNoAnonym;       // don't allow an anonymous login
    sFtpServer.iPriority        = pConfig->iPriority;       // taks priority for client connections
    sFtpServer.iMaxConnections  = pConfig->iMaxConnections; // maximum count of  client connections
    sFtpServer.usPassivePort    = FTP_PASSIVE_MIN;          // start port for passive connections
    sFtpServer.iCurConnections  = 0;                        // current count of  client connections

    if(!pConfig->iMaxConnections)
        {
        pConfig->iMaxConnections = 16;
        }

    if(!pConfig->usTcpPort)
        {
        pConfig->usTcpPort = 21;
        }

    if(pConfig->uTimeout<5000)
        {
        pConfig->uTimeout = 5000;
        }

    sFtpServer.iRunning         = 1;
    sFtpServer.iReadOnly        = pConfig->iReadOnly;
    sFtpServer.uTimeout         = pConfig->uTimeout;
    sFtpServer.pCbLogin         = pConfig->pCbLogin;
    sFtpServer.pCbConnect       = pConfig->pCbConnect;
    sFtpServer.pCbDisconnect    = pConfig->pCbDisconnect;
    sFtpServer.pCbChecker       = pConfig->pCbChecker;
    sFtpServer.pCbCheckerParam  = pConfig->pCbCheckerParam;

    strncpy(sFtpServer.cFilePath ,pConfig->cFilePath ,sizeof(sFtpServer.cFilePath ));
    strncpy(sFtpServer.cIpAddress,pConfig->cIpAddress,sizeof(sFtpServer.cIpAddress));


                                                            // start listen thread
    if(!SysThreadStart(FtpListenProc,512,&sFtpServer,sFtpServer.iPriority|SYS_GLOBAL,"FtpListen"))
        {
        SysSemaphoreLock(sFtpServer.hSemaphore);
        TcpClose(sFtpServer.hServer);
        sFtpServer.hServer = NULL;
        SysSemaphoreUnlock(sFtpServer.hSemaphore);

        SysSemaphoreDelete(sFtpServer.hSemaphore);
        sFtpServer.hSemaphore = NULL;

        return FTP_ERR_INIT;
        }


    iServerRunning = 1;



return FTP_ERR_NONE;
}

//*****************************************************************************
//*
//*     FtpStopDaemon
//*
//*****************************************************************************
//  Stop the http server
//  Returns 0 it daemon was started, or an error code greather zero
//                  FTP_ERR_NONE         0   no error
//                  FTP_ERR_ISNOTRUNNING 3   ftp server is not running
//                  FTP_ERR_INIT         5   initialization error
//                  FTP_ERR_NOTSTOPED    6   the server could not stopped
FTP_API int FtpStopDaemon()
{
int     iTurn;



    if(!iServerRunning)                                     // is the server allready running
        {
        return FTP_ERR_ISNOTRUNNING;
        }

    SysSemaphoreLock(sFtpServer.hSemaphore);
    sFtpServer.iShutdown = 1;
    if(sFtpServer.hServer)TcpShutdown(sFtpServer.hServer);
    SysSemaphoreUnlock(sFtpServer.hSemaphore);

    for(iTurn=0;;iTurn++)                                   // wait until server was stopped
        {
        if(iTurn>10*20)return FTP_ERR_NOTSTOPED;
        if(!sFtpServer.iRunning)
        if(!sFtpServer.iCurConnections)break;
        SysSleep(100);
        }


    iServerRunning = 0;


return FTP_ERR_NONE;
}

//*****************************************************************************
//*
//*     FtpInsertTreeEntry
//*
//*****************************************************************************
//  Inserts a share to a FTP connection. This function is used in
//  the login callback function.
//  hConn       : Handle of the connection.
//  pName       : Is the name of the share. Use "" for a default share with the
//                base path "". Only one default share is allowed.
//  pPath       : Is the path of the share. (i.e "C:\MyDir" or "C:\MyDir\")
//  iReadOnly   : Is the share in the read only mode
//  pCheckCb    : Is the callback for checking file access
//  pParam      : Is the user parameter for the checking callback
//  Returns 0 if the share was inserded or an erro code.
//                  FTP_ERR_NONE            0   no error
//                  FTP_ERR_DOMUCHSHARES    8   too much shares defined
//                  FTP_ERR_WRONGNAME       9   wrong share name
//					FTP_ERR_DOUBLENAME     11   double share name 
FTP_API int FtpInsertTreeEntry(FtpConnection hConn,const char *pName,const char *pPath,int iReadOnly,FtpFileCheckerCB pCheckCb,void *pParam)
{
FtpData        *pData=(FtpData*)hConn;
FtpShare       *pShare;
int             iShare;
int             iLen;



    if(pData->iShareCount>=FTP_SHARES)                      // do much shares defined
        {
        return FTP_ERR_DOMUCHSHARES;
        }


     for(iShare=0;iShare<pData->iShareCount;iShare++)	    // exist a share with the same name
        {
        if(pData->aShares[iShare].cName[0]==0)continue;
		if(strcmp(pData->aShares[iShare].cName,pName))continue;
        return FTP_ERR_DOUBLENAME;
        }


    pShare  = pData->aShares+pData->iShareCount;
    iLen    = strlen(pName);

    if(iLen>=FTP_MAX_NAME)
        {
        return FTP_ERR_WRONGNAME;
        }


    memcpy(pShare->cName,pName,iLen+1);
    pShare->iNameLen = iLen;


    iLen    = strlen(pPath);

    if(iLen>=FTP_MAX_PATH-1)
        {
        return FTP_ERR_WRONGNAME;
        }


    memcpy(pShare->cPath,pPath,iLen);
                                                            // add a backslash
    if(iLen<1 || (pPath[iLen-1]!='\\' && pPath[iLen-1]!='/'))
        {
        pShare->cPath[iLen] = '/';
        iLen++;
        }

    pShare->cPath[iLen] =  0;
    pShare->iPathLen    =  iLen;
    pShare->cReadOnly   = (iReadOnly)? 1:0;
    pShare->pCbChecker  =  FTP_CHECKER(pCheckCb);
    pShare->pCbParam    =  pParam;


    for(iLen--;iLen>=0;iLen--)                              // convert \ to /
        {
        if(pShare->cPath[iLen]!='\\')continue;
        pShare->cPath[iLen]='/';
        }


    if(!pShare->iNameLen)
        {
        pData->iShareRoot = pData->iShareCount;
        }

    pData->iShareCount++;



return FTP_ERR_NONE;
}


//*****************************************************************************
//*
//*     FtpGetUserValue
//*
//*****************************************************************************
//  Gets an user value
//  hConn       : pointer to the connection data
//  uNumber     : number of the uservalue
//  Returns the user value
FTP_API int FtpGetUserValue(FtpConnection hConn,unsigned uNumber)
{
FtpData   *pData=(FtpData*)hConn;


    if(uNumber>=FTP_USER_PARAMS)return 0;

return (int)pData->pUserParams[uNumber];
}

//*****************************************************************************
//*
//*     FtpGetUserPointer
//*
//*****************************************************************************
//  Gets an user pointer
//  hConn       : pointer to the connection data
//  uNumber     : number of the uservalue
//  Returns the user pointer
FTP_API void *FtpGetUserPointer(FtpConnection hConn,unsigned uNumber)
{
FtpData   *pData=(FtpData*)hConn;

    if(uNumber>=FTP_USER_PARAMS)return 0;

return pData->pUserParams[uNumber];
}

//*****************************************************************************
//*
//*     FtpSetUserValue
//*
//*****************************************************************************
//  Sets an user value
//  hConn       : pointer to the connection data
//  uNumber     : number of the uservalue
//  iValue      : is the new value
FTP_API void FtpSetUserValue(FtpConnection hConn,unsigned uNumber,int iValue)
{
FtpData   *pData=(FtpData*)hConn;

    if(uNumber>=FTP_USER_PARAMS)return;

    pData->pUserParams[uNumber] = (void*)iValue;
}

//*****************************************************************************
//*
//*     FtpSetUserValue
//*
//*****************************************************************************
//  Sets an user pointer
//  hConn       : pointer to the connection data
//  uNumber     : number of the uservalue
//  pPointer    : is the new pointer
FTP_API void FtpSetUserPointer(FtpConnection hConn,unsigned uNumber,void *pPointer)
{
FtpData   *pData=(FtpData*)hConn;

    if(uNumber>=FTP_USER_PARAMS)return;

    pData->pUserParams[uNumber] = pPointer;

}

